/******************************************************************************
 * Module Name : $RCSfile: CoverDate.java,v $
 *
 * Description : CoverDate, for an issue.
 *
 * Updates:
 *   $Log: CoverDate.java,v $
 *   Revision 1.5  2001/07/29 00:47:29  jcouper
 *   job:137
 *   Added support for credits, and re-vamped the EditDialog.
 *
 *   Revision 1.4  2000/02/18 22:35:35  Jonathan
 *   Add basic application settings support.
 *
 *   Revision 1.3  2000/01/10 02:22:32  Jonathan
 *   Add new get/set methods.
 *
 *   Revision 1.2  2000/01/07 05:43:34  Jonathan
 *   Render fixed cell dates differently
 *
 *   Revision 1.1  1999/10/25 08:37:51  Jonathan
 *   Class to track CoverDates.
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.util.*;

public class CoverDate {
    /*
     * true  - This is a definitive date, from which subsequent dates are computed.
     * false - This is a derived date.
     */
    boolean mark = false;

    /*
     * Is this date actually defined?
     */
    boolean defined = false;

    /*
     * Year, Month
     */
    int year = 0;               // 1900-2100 is valid if defined is true.
    double month = 0.0;         // 1-12 is valid if defined = true.

    static final public int YEAR_MIN=1900;
    static final public int YEAR_MAX=2099;

    static final public int STEP_SINGLETON = 0;
    static final public int STEP_WEEKLY = 1;
    static final public int STEP_TWO_WEEKLY = 2;
    static final public int STEP_MONTHLY = 3;
    static final public int STEP_TWO_MONTHLY = 4;
    static final public int STEP_THREE_MONTHLY = 5;
    static final public int STEP_SIX_MONTHLY = 6;
    static final public int STEP_ANNUAL = 7;
    static final public int STEP_MAX = 7;

    double stepMonths[] = {0.0, 0.25, 0.5, 1.0, 2.0, 3.0, 6.0, 12.0};
    static final public String monthCodes[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                                    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", "???"};

    static final public String stepStrings[] = {"One Shot", "Weekly", "Two-Weekly", "Monthly",
                                   "Two-Monthly", "Quarterly", "Bi-Annual", "Annual"};

    int step = 0;

    /*
     **************************************************************************
     * Constructors.
     **************************************************************************
     */
    /*
     * Generate a new undefined cover date.
     */
    public CoverDate () {
        defined = false;
    }

    /*
     * Generate a new definitive cover date, defined to the indicated parameters.
     */
    public CoverDate (int yearIn, int monthIn, int stepIn) {
        configure (yearIn, monthIn, stepIn);
    }

    /*
     **************************************************************************
     * Override our inherited methods.
     **************************************************************************
     */
    /*
     * Return ourself as a string.  More formatting options are required here.
     */
    public String toString () {
        String result = null;

        if (defined) {
            if (isValid()) {
                result = String.valueOf (year) + " " + monthCodes[(int) month - 1];
            } else {
                result = "Bad Data";
            }

        } else {
            result = "";
        }

        return result;
    }

    /*
     **************************************************************************
     * Get and Set methods.
     **************************************************************************
     */
    public boolean isMark () {
        return mark;
    }
    public boolean isDefined () {
        return defined;
    }
    public void setMark (boolean newMark) {
        mark = newMark;
    }
    public void setDefined (boolean newDefined) {
        defined = newDefined;
    }

    public int getYear () {
        return year;
    }
    public int getMonth () {
        return (int) month;
    }
    public int getStep () {
        return step;
    }
    public void setYear (int newYear) {
        year = newYear;
    }
    public void setMonth (int newMonth) {
        month = newMonth;
    }
    public void setStep (int newStep) {
        if ((newStep < STEP_SINGLETON) || (newStep > STEP_MAX)) {
            return;
        }
        step = newStep;
    }

    /*
     **************************************************************************
     * CONFIGURATION METHODS.
     **************************************************************************
     */
    /*
     * Configure our own parameters.
     */
    public void configure (int yearIn, int monthIn, int stepIn) {
        mark = true;
        defined = true;

        step = stepIn;
        year = yearIn;
        month = (double) monthIn;

        if (!isValid()) {
            this.defined = false;
        }
    }

    /*
     * Configure a cover date, stepped from a previous one.
     */
    public void configureSubsequent (CoverDate subsequent) {
        // Don't trample a mark date!
        if (subsequent.mark) {
            return;
        }
        // If we're not defined or valid, we can't derive from ourself.
        if (!defined) {
            subsequent.defined = false;
            return;
        }
        if (!isValid()) {
            subsequent.defined = false;
            return;
        }
        // Singleton's can't be used for subsequent derivation.
        if (step == STEP_SINGLETON) {
            subsequent.defined = false;
            return;
        }
        // Now step from ourself onwards.
        subsequent.defined = true;
        subsequent.mark = false;
        subsequent.step = step;
        subsequent.year = year;
        subsequent.month = month;

        // Annual, we do the year explicitly, to cope with the
        // month = 13 (unspecified) case, which will break otherwise.
        if (step == STEP_ANNUAL) {
            subsequent.year = subsequent.year + 1;
        } else {
            subsequent.month = subsequent.month + stepMonths[step];

            while (subsequent.month >= 13) {
                subsequent.year = subsequent.year + 1;
                subsequent.month = subsequent.month - 12.0;
            }
        }
        return;
    }

    /*
     * Configure a mark date from a "yyyymmdd [step]" string.
     */
    public void configure (String parameters)
    {
        StringTokenizer tokenizer = new StringTokenizer (parameters);

        // No token?  That's bad.
        if (!tokenizer.hasMoreTokens()) {
            defined = false;
            return;
        }

        // Great, decode the yyyymm string.
        String yyyymmString = tokenizer.nextToken();
        if (yyyymmString.length() != 6) {
            defined = false;
            return;
        }
        year = (new Integer (yyyymmString.substring (0, 4))).intValue();
        month = (new Integer (yyyymmString.substring (4, 6))).intValue();

        // Then the optional step string.
        if (tokenizer.hasMoreTokens()) {
            String stepString   = tokenizer.nextToken();
            step = (new Integer (stepString)).intValue();

        } else {
            step = STEP_MONTHLY;
        }

        // We're now a valid mark date...
        defined = true;
        mark = true;

        // ... aren't we?
        if (!isValid()) {
            defined = false;
            return;
        }
    }

    /*
     * Perform our bounds checking.
     */
    public boolean isValid () {
        if (!defined) {
            return true;
        }
        if ((step < 0) || (step > STEP_MAX)) {
            return false;
        }
        if ((year < YEAR_MIN) || (year > YEAR_MAX)) {
            return false;
        }
        // Allow month 13, it's our no-month case, valid only for
        // single or annual covers.
        if ((step == STEP_SINGLETON) || (step == STEP_ANNUAL)) {
            if ((month < 1.0) || (month > 13.0)) {
                return false;
            }
        } else {
            if ((month < 1.0) || (month >= 13.0)) {
                return false;
            }
        }
        return true;
    }

    /*
     **************************************************************************
     * ADDITIONAL FORMAT METHODS.
     **************************************************************************
     */
    /*
     * Return the parameters to a /Date line which we could put to a template
     * file before the issue which had this cover date.
     */
    public String getDateParameters () {
        if (!mark) {
            return null;
        }

        /*
         * A mark which is not defined, or not valid?
         * Terminate the date sequencing.
         */
        if (!defined || !isValid()) {
            return "";
        }

        return lpad (year, 4) + lpad ((int) month, 2) + " " + step;
    }

    protected String lpad (int value, int count) {
        String string = String.valueOf (value);
        while (string.length() < count) {
            string = "0" + string;
        }
        return string;
    }
}