/******************************************************************************
 * Module Name : $RCSfile: Index.java,v $
 *
 * Description : Index of all collections.  Represents the contents of
 *                an index.dat file.
 *
 * Updates:
 *   $Log: Index.java,v $
 *   Revision 1.10  2001/04/11 08:22:24  jcouper
 *   Add Filter Support
 *
 *   Revision 1.9  2001/04/04 07:34:53  jcouper
 *   Support credits, filters, new columns, etc, etc.  Stable release.
 *
 *   Revision 1.8  2000/08/23 19:37:06  jcouper
 *   Add sort titles by name method.
 *
 *   Revision 1.7  2000/03/18 00:46:13  Jonathan
 *   Big changes in the way the IndexDialog works, including a new CardDialog.
 *   New preferences for column display.  Permit renaming of title files.
 *
 *   Revision 1.6  2000/02/19 00:18:53  Jonathan
 *   Add saving of settings.
 *
 *   Revision 1.5  1999/10/25 08:33:07  Jonathan
 *   v0_9_2
 *
 *   Revision 1.4  1999/10/24 22:37:55  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.util.*;
import java.io.*;
import standup.*;
import standup.util.*;

public class Index extends Vector {
    // Application Settings
    ApplicationSettings applicationSettings = Application.getSettings ();

    // Where the data came from, for later reference, e.g. when writing.
    File indexFile = null;

    // Our internal static list of all known titles.
    Vector allTitles = null;

    // Have we been modified since loading?
    boolean modified = false;

    /*
     **************************************************************************
     * CONSTRUCTOR
     **************************************************************************
     * Create an Index from the contents of an index file.
     */
    public Index (File newIndexFile) {
        /*
         * Making a new index?  That's going to mean a title refresh.
         */
        indexFile = newIndexFile;

        /*
         * Hey, better have template and user directories.
         */
        File templateDir = new File (indexFile.getParentFile(), "template/");
        if (!templateDir.exists() && !templateDir.mkdirs()) {
            System.out.println ("Index: Cannot make " + templateDir);
            return;
        }
        File userDir = new File (indexFile.getParentFile(), "user/");
        if (!userDir.exists() && !userDir.mkdirs()) {
            System.out.println ("Index: Cannot make " + userDir);
            return;
        }

        initAllTitles ();

        /*
         * Read the index file to see the contents.  Read the help file
         * for a definition of the contents and syntax of the index file.
         */
        Collection currentCollection = null;
        Card currentCard = null;
        TitleRef currentTitleRef = null;

        try {
            BufferedReader input = new BufferedReader (new FileReader (indexFile));

            String line = null;

            while ((line = input.readLine()) != null) {
                // Trim lines which start with #
                if ((line.length() == 0) || (line.charAt(0) == '#')) {
                    continue;
                }

                // Get first token
                MyStringTokenizer tokenizer = new MyStringTokenizer (line.trim());
                String firstToken = null;

                try {
                    firstToken = tokenizer.nextToken();
                }
                catch (NoSuchElementException nse) {
                    continue;
                }

                /*
                 * /[Cc]ollection <name>
                 * Add a collection to the collection list.
                 */
                if (firstToken.toLowerCase().equals("/collection")) {
                    String collectionName = tokenizer.remainder().trim();
                    if (collectionName.equals("")) {
                        System.out.println ("Missing Collection Name in '" + indexFile + "'");
                        System.out.println ("> " + line);
                        continue;
                    }
                    currentCollection = new Collection (this, collectionName);
                    this.addElement (currentCollection);

                /*
                 * /[Cc]ard <name>
                 * Add a card to the current collection.
                 */
                } else if (firstToken.toLowerCase().equals("/card")) {
                    String cardName = tokenizer.remainder().trim();
                    if (cardName.equals("")) {
                        System.out.println ("Missing Card Name in '" + indexFile + "'");
                        System.out.println ("> " + line);
                        continue;
                    }
                    currentCard = new Card (cardName);
                    currentCollection.addElement (currentCard);

                /*
                 * /[Tt]itleRef <name>
                 * Add a title reference to the current card.
                 */
                } else if (firstToken.toLowerCase().equals("/titleref")) {
                    try {
                        String titleRefKeyName = tokenizer.nextToken();
                        String refName = tokenizer.remainder().trim();
                        if (!refName.equals("")) {
                            currentTitleRef = new TitleRef (this, titleForKeyname (titleRefKeyName), refName);

                        } else {
                            currentTitleRef = new TitleRef (this, titleForKeyname (titleRefKeyName));
                        }
                        currentCard.addElement (currentTitleRef);
                    }
                    catch (NoSuchElementException nse) {
                        System.out.println ("Missing Title Ref Name in '" + indexFile + "'");
                        System.out.println ("> " + line);
                        continue;
                    }

                /*
                 * /[Ff]ilter <string>
                 * Set filter for preceeding TitleRef..
                 */
                } else if (firstToken.toLowerCase().equals("/filter")) {
                    String filterString = tokenizer.remainder().trim();
                    if (filterString.equals("")) {
                        System.out.println ("Missing Filter String in '" + indexFile + "'");
                        System.out.println ("> " + line);
                        continue;
                    }
                    currentTitleRef.setFilterFromString (filterString);
                }
            }
            input.close ();
        }
        catch (FileNotFoundException fnfe) {
            // System.out.println ("File not found: Creating now.");
            try {
                indexFile.createNewFile();
            }
            catch (IOException io) {
                System.out.println ("Error creating file: " + indexFile);
            }
        }
        catch (IOException io) {
            System.out.println ("Error reading file: " + indexFile);
            io.printStackTrace();
        }

        /*
         * No Collection?  Better Create one!
         */
        if (size() <= 0) {
            Collection collection = new Collection (this, "Sample Collection");
            Card card = new Card ("Sample Card");

            collection.addElement (card);
            addElement (collection);
        }

        /*
         * Hey, it worked, this is our new last opened index.
         */
        applicationSettings.lastIndex = indexFile.getPath();
    }

    /*
     **************************************************************************
     * Override our inherited methods.
     **************************************************************************
     */
    public void addElement (Object object) {
        if (! (object instanceof Collection)) {
            System.out.println ("Index is Vector of Collection");
            System.exit (1);
        }
        super.addElement (object);
    }

    public String toString () {
        if (indexFile == null) {
            return "New Index";
        } else {
            return "Index (" + indexFile.getName() + ")";
        }
    }

    /*
     **************************************************************************
     * Get and set methods.
     **************************************************************************
     * Modified flag for explicitly.
     */
    public void setModified (boolean newModified) {
        modified = newModified;
    }

    public boolean isModified () {
        if (modified) {
            return true;

        } else {
            for (Enumeration e = allTitles.elements() ; e.hasMoreElements() ;) {
                Title title = (Title) e.nextElement();
                if (title.isRenameRequired()) {
                    return true;
                }
            }
        }
        return false;
    }

    /*
     * Be nice with allTitles.  May modifify this to return a clone, just
     * to make sure you don't break it.
     */
    public Vector getAllTitles () {
        return allTitles;
    }

    public File getIndexFile () {
        return indexFile;
    }

    /*
     **************************************************************************
     * OTHER NIFTY STUFF
     **************************************************************************
     */
    /*
     * This will give you a title for the keyname, it will give you an existing
     * one if it exists, or a brand new one otherwise.
     */
    public Title titleForKeyname (String newKeyname) {
        // Try an existing one.
        Title existingTitle = existingTitleForKeyname (newKeyname);
        if (existingTitle != null) {
            return existingTitle;
        }

        // Nothing yet?  Create a new one.
       return newTitleForKeyname (newKeyname);
    }

    /*
     * This one will only give you existing titles from our list.
     */
    public Title existingTitleForKeyname (String newKeyname) {
       for (Enumeration e = allTitles.elements() ; e.hasMoreElements() ;) {
            Title title = (Title) e.nextElement();
            if (title.getKeyname().equals(newKeyname)) {
                return title;
            }
       }
        return null;
    }

    /*
     * This one will give you a brand new one, except that it fails and
     * returns null if that keyname is already taken.
     */
    public Title newTitleForKeyname (String newKeyname) {
        Title existingTitle = existingTitleForKeyname (newKeyname);
        if (existingTitle != null) {
            return null;
        }

        File templateFile = new File (indexFile.getParentFile(), "template/" + newKeyname + ".tem");
        File userFile = new File (indexFile.getParentFile(), "user/" + newKeyname + ".dat");
        Title newTitle = new Title (newKeyname, templateFile, userFile);

        allTitles.addElement (newTitle);
        return newTitle;
    }

    public boolean containsReferenceTo (Title title) {
        for (Enumeration e = elements() ; e.hasMoreElements() ;) {
            Collection collection = (Collection) e.nextElement();

            for (Enumeration e2 = collection.elements() ; e2.hasMoreElements() ;) {
                Card card = (Card) e2.nextElement();

                for (Enumeration e3 = card.elements() ; e3.hasMoreElements() ;) {
                    TitleRef titleRef = (TitleRef) e3.nextElement();

                    if (titleRef.getTitle().equals (title)) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /*
     * OK, this is the actual list of all titles.  Be nice to it.  Break it, and
     * things will go bad.  Initialise us first, please.
     */
    private void initAllTitles () {
        File templateDir = new File (indexFile.getParent(), "/template/");
        File titleList[] = templateDir.listFiles();

        allTitles = new Vector();
        if (titleList == null) {
            // System.out.println ("initAllTitles: No template dir exists.");

        } else {
            for (int i=0; i<titleList.length; i++) {
                String name = titleList[i].getName();
                if (name.endsWith (".tem")) {
                    String nameWithoutTem = name.substring (0, name.length() - 4);
                    newTitleForKeyname (nameWithoutTem);
                }
            }
        }
        sortAllTitlesByGetAnyName ();
    }

    /*
     * Good old bubble sort of all titles by getName(), A..Z.
     */
    public void sortAllTitlesByGetAnyName () {
        boolean sorting = true;
        while (sorting) {
            sorting = false;

            for (int i=0; i<allTitles.size()-1; i++) {
                Title firstTitle = (Title) allTitles.elementAt (i);
                Title secondTitle = (Title) allTitles.elementAt (i+1);

                if (firstTitle.getAnyName().compareToIgnoreCase(secondTitle.getAnyName()) > 0) {
                    allTitles.setElementAt (secondTitle, i);
                    allTitles.setElementAt (firstTitle, i+1);
                    sorting = true;
                }
            }
        }
    }

    /*
     * Good old bubble sort of collections by getName(), A..Z.
     */
    public void sortCollectionsByName () {
        boolean sorting = true;
        while (sorting) {
            sorting = false;

            for (int i=0; i<size()-1; i++) {
                Collection firstCollection = (Collection) elementAt (i);
                Collection secondCollection = (Collection) elementAt (i+1);

                if (firstCollection.getName().compareToIgnoreCase(secondCollection.getName()) > 0) {
                    setElementAt (secondCollection, i);
                    setElementAt (firstCollection, i+1);
                    sorting = true;
                }
            }
        }
    }

    /*
     **************************************************************************
     * SAVE METHOD
     **************************************************************************
     * This method saves the index back to the file it came from.  As part of
     * this, if we have any file renames which were pending, we do these first.
     */
    public void save () {
        /*
         * Perform all our pending renames.  We do this BEFORE we save, so that
         * the index we are now saving will have all of the NEW names, not the
         * old ones.
         */
        for (Enumeration e = allTitles.elements() ; e.hasMoreElements() ;) {
            Title title = (Title) e.nextElement();
            title.renameIfRequired();
        }

        try {
            BufferedWriter output = new BufferedWriter (new FileWriter (indexFile));

            output.write ("#");
            output.newLine ();
            output.write ("# This file " + indexFile + " generated automatically");
            output.newLine ();
            output.write ("# by Standup Comic Management Software (Version " + Application.getVersion() + ")");
            output.newLine ();
            output.write ("#");
            output.newLine ();

            for (Enumeration e = elements() ; e.hasMoreElements() ;) {
                Collection collection = (Collection) e.nextElement();
                output.write ("/Collection " + collection.getName());
                output.newLine ();

                for (Enumeration e2 = collection.elements() ; e2.hasMoreElements() ;) {
                    Card card = (Card) e2.nextElement();
                    output.write ("/Card " + card.getName());
                    output.newLine ();

                    for (Enumeration e3 = card.elements() ; e3.hasMoreElements() ;) {
                        TitleRef titleRef = (TitleRef) e3.nextElement();
                        String refName = titleRef.getName();

                        if ((refName != null) && !refName.equals("")) {
                            output.write ("/TitleRef " + titleRef.getTitle().getKeyname() + " " + refName);

                        } else {
                            output.write ("/TitleRef " + titleRef.getTitle().getKeyname());
                        }
                        output.newLine ();

                        String filterString = titleRef.getFilterAsString ();
                        if ((filterString != null) && !filterString.equals("")) {
                            output.write ("/Filter " + filterString);
                            output.newLine ();
                        }
                    }
                }
                output.newLine ();
            }

            output.close ();
            modified = false;
        }
        catch (IOException io) {
            System.out.println ("Error writing file: " + indexFile);
        }
    }
}