/******************************************************************************
 * Module Name : $RCSfile: Issue.java,v $
 *
 * Description : Issue, belongs to a title.
 *
 * Updates:
 *   $Log: Issue.java,v $
 *   Revision 1.20  2003/09/09 22:30:28  jcouper
 *   161: Add support for / Info of any form
 *
 *   Revision 1.19  2003/01/23 06:37:46  jcouper
 *   Version 1.1 Changes
 *
 *   Revision 1.18  2002/07/14 03:48:31  jcouper
 *   job:159
 *   Option to over-ride the automatic value.
 *
 *   Revision 1.17  2002/07/13 02:22:29  jcouper
 *   job:157
 *   Still tweaking after changing grading and valuation.
 *
 *   Revision 1.16  2002/07/13 01:06:17  jcouper
 *   job:157
 *   Add +/- for grades.  (Also fixed 147 at the same time).
 *
 *   Revision 1.15  2001/10/11 21:28:04  jcouper
 *   142: Added _ALT option to Issue menu
 *
 *   Revision 1.14  2001/07/29 00:47:29  jcouper
 *   job:137
 *   Added support for credits, and re-vamped the EditDialog.
 *
 *   Revision 1.13  2001/07/28 21:53:34  jcouper
 *   Added ON ORDER code
 *
 *   Revision 1.12  2001/04/04 07:34:53  jcouper
 *   Support credits, filters, new columns, etc, etc.  Stable release.
 *
 *   Revision 1.11  2000/08/23 19:38:05  jcouper
 *   Add method to allow update of non-numeric issue names.  This means we
 *   won't lose all our other data if we change an issue name.  Yay!
 *
 *   Revision 1.10  2000/02/19 00:18:53  Jonathan
 *   Add saving of settings.
 *
 *   Revision 1.9  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.8  2000/01/07 05:43:34  Jonathan
 *   Render fixed cell dates differently
 *
 *   Revision 1.7  1999/11/28 07:13:40  Jonathan
 *   Added Flags to Titles, and Polished AllTitlesMatrix report
 *
 *   Revision 1.6  1999/10/25 08:33:07  Jonathan
 *   v0_9_2
 *
 *   Revision 1.5  1999/10/24 22:37:55  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.util.*;
import standup.*;

public class Issue {
    // Application Settings
    ApplicationSettings applicationSettings = Application.getSettings ();

    // How many credits do we permit per issue.
    static public final int MAX_CREDITS = 6;

    // These are the grade index numbers I recognise.
    static public final int GRADE_NONE = 0;
    static public final int GRADE_UNKNOWN = 1;
    static public final int GRADE_ON_ORDER = 2;
    static public final int GRADE_REPRINT = 3;
    static public final int GRADE_POOR = 4;
    static public final int GRADE_FAIR = 5;
    static public final int GRADE_GOOD = 6;
    static public final int GRADE_VERY_GOOD = 7;
    static public final int GRADE_FINE = 8;
    static public final int GRADE_VERY_FINE = 9;
    static public final int GRADE_NEAR_MINT = 10;
    static public final int GRADE_MINT = 11;
    static public final int GRADE_MAX = 11;

    // And we can adjust up or down... -- to ++
    static public final int GRADE_MOD_PLUS_PLUS = 2;
    static public final int GRADE_MOD_PLUS = 1;
    static public final int GRADE_MOD_NONE = 0;
    static public final int GRADE_MOD_MINUS = -1;
    static public final int GRADE_MOD_MINUS_MINUS = -2;

    // These are some nice names for grade numbers.
    static private final String gradeAbbreviations[] = {
        "<none>", "<???>", "<ord>", "<rep>", "PR", "FR", "G", "VG", "FN", "VF", "NM", "M"};

    static private final String gradeNames[] = {
        "Not Held", "Unknown", "On Order", "Reprint", "Poor", "Fair", "Good", "Very Good",
        "Fine", "Very Fine", "Near Mint", "Mint"};

    static public final int gradeDefaultPercents[] = {
        0, 0, 0, 0, 3, 5, 12, 25, 40, 70, 100, 120};

    /*
     * Issue details.
     */
    String code = null;
    double nmValue = 0.0;
    double gradeValue = 0.0;    // 0.0 means use NM Value * Grade
    int grade = GRADE_NONE;
    int gradeMod = GRADE_MOD_NONE;
    CoverDate coverDate = new CoverDate();
    String info = null;
    String extra = null;
    String comment = null;
    Credit[] credits = new Credit[MAX_CREDITS];
    int numCredits = 0;

    // Handy rounding function
    private double roundCents (double cents, double nearest) {
        return (nearest * (int) ((cents + nearest/2.0) / nearest));
    }

    /*
     **************************************************************************
     * Constructors.
     **************************************************************************
     */
    /*
     * Absolute simplest constructor.
     */
    public Issue (String newCode) {
        code = newCode;
    }

    /*
     * Use this constructor if you get the info from a template file.
     */
    public Issue (String newCode, double newNmValue, String newInfo) {
        code = newCode;
        nmValue = newNmValue;
        info = newInfo;
    }

    /*
     * Full constructor.
     */
    public Issue (String newCode, double newNmValue,
                  int newGrade, int newGradeMod,
                  String newInfo, String newComment) {

        code = newCode;
        nmValue = newNmValue;
        grade = newGrade;
        if (grade <= GRADE_REPRINT) {
            gradeMod = GRADE_MOD_NONE;

        } else {
            gradeMod = newGradeMod;
        }
        info = newInfo;
        comment = newComment;
    }

    /*
     * Initialise user data, as from user data file.
     */
    public void initUserData (String newGradeString, String newComment) {

        setGradeAndMod (newGradeString);
        comment = newComment;
    }

    /*
     * Use these to set the cover date when you have an explicit /Date on the
     * previous line.
     */
    public void setCoverDateExplicit (String parameters) {
        coverDate.configure (parameters);
    }

    /*
     * Generate a new cover date, stepped from the previous one.
     */
    public void setCoverDateFromPrevious (Issue previous) {
        previous.coverDate.configureSubsequent (this.coverDate);
    }

    /*
     * Generate a new cover date, exactly equal to the previous one.
     */
    public void copyCoverDate (Issue source) {
        coverDate.configure (source.coverDate.getYear(),
                             source.coverDate.getMonth(),
                             source.coverDate.getStep());
    }

    /*
     **************************************************************************
     * Get and set TEMPLATE PARAMETERS.
     **************************************************************************
     */
    public String getCode () {
        return code;
    }

    // This works IFF the old and new codes are both non-numeric.
    public void setCode (String newCode) {
        if (isCodeNumeric()) {
            System.out.println ("Issue: Cannot update numeric code");
            return;
        }
        try {
            Integer codeAsInt = new Integer (newCode);
            System.out.println ("Issue: Cannot update code to numeric value");
            return;
        }
        catch (NumberFormatException nfe) {
            code = newCode;
        }
        return;
    }

    public Integer getCodeAsInteger () {
        try {
            Integer codeAsInt = new Integer (code);
            return codeAsInt;
        }
        catch (NumberFormatException nfe) {
            return null;
        }
    }

    public int getCodeAsInt () {
        Integer codeAsInteger = getCodeAsInteger();
        if (codeAsInteger == null) {
            return 0;

        } else {
            return codeAsInteger.intValue();
        }
    }

    public boolean isCodeNumeric () {
        return (getCodeAsInteger() != null);
    }

    public double getNmValue () {
        return nmValue;
    }

    public void setNmValue (double newNmValue) {
        nmValue = newNmValue;
    }


    public double getAutoGradeValue () {
        if ((grade < 0) || (grade > GRADE_MAX)) {
            return 0.0;
        }
        if (nmValue <= 0) {
            return 0.0;
        }

        // Work out what percent of grade this is worth
        double gradePercent;
        if (gradeMod == GRADE_MOD_NONE) {
            gradePercent = applicationSettings.gradePercentValues[grade];

        // Positive grades mods - can't get better than Mint
        } else if (gradeMod > GRADE_MOD_NONE) {
            if (grade >= GRADE_MINT) {
                gradePercent =  applicationSettings.gradePercentValues[GRADE_MINT];

            } else {
                gradePercent = applicationSettings.gradePercentValues[grade] +
                    0.2 * gradeMod *
                       (applicationSettings.gradePercentValues[grade+1]
                        - applicationSettings.gradePercentValues[grade]);
            }

        // Negative grades mods - can't do worse than Poor
        } else {
            if (grade <= GRADE_POOR) {
                gradePercent = applicationSettings.gradePercentValues[GRADE_POOR];

            } else {
                gradePercent = applicationSettings.gradePercentValues[grade] +
                    0.2 * gradeMod *
                       (applicationSettings.gradePercentValues[grade]
                        - applicationSettings.gradePercentValues[grade-1]);
            }
        }

	double cents = gradePercent * nmValue;

	// Nothing is still Nothing
        if (cents <= 0.0) {

    	// Below $2, we round to nearest 25c.
        } else if (cents <= 200.0) {
            cents = roundCents (cents, 25.0);
	    if (cents <= 0.0) {
		cents = 25.0;
	    }

    	// Below $10, we round to nearest 50c.
        } else if (cents <= 1000.0) {
            cents = roundCents (cents, 50.0);
	
	// Below $20, we round to nearest $1.
        } else if (cents <= 2000.0) {
            cents = roundCents (cents, 100.0);
	
	// Below $150, we round to nearest $5.
        } else if (cents <= 15000.0) {
            cents = roundCents (cents, 500.0);
	
	// Below $1000, we round to nearest $10
        } else if (cents <= 100000.0) {
            cents = roundCents (cents, 1000.0);
	
	// Over, nearest $50
        } else {
            cents = roundCents (cents, 5000.0);
        }	

	// Return as dollars
        return (cents / 100.0);
    }

    public CoverDate getCoverDate () {
        return coverDate;
    }

    public String getDateParameters () {
        return coverDate.getDateParameters();
    }

    public String getInfo () {
        return (info == null) ? "" : info;
    }

    public void setInfo (String newInfo) {
        if ((info == null) && (newInfo == null)) {
            return;
        }
        if ((info != null) && (newInfo != null) && info.equals(newInfo)) {
            return;
        }
        info = newInfo;
    }

    public String getExtra () {
        return (extra == null) ? "" : extra;
    }

    public void setExtra (String newExtra) {
        if ((extra == null) && (newExtra == null)) {
            return;
        }
        if ((extra != null) && (newExtra != null) && extra.equals(newExtra)) {
            return;
        }
        extra = newExtra;
    }
    
    public void addExtra (String moreExtra) {
        if (extra == null) {
            extra = moreExtra;
        
	} else {
	    extra = extra + "\n" + moreExtra;
        }
    }

    public boolean isHeld () {
        return ((grade != GRADE_NONE) && (grade != GRADE_ON_ORDER));
    }

    public void addCredit (String what, String who) {
        if (numCredits >= MAX_CREDITS) {
            System.out.println ("WARNING: Overflow credits (exceed " + MAX_CREDITS + ") for code " + code);
            return;
        }

        credits[numCredits] = new Credit (what, who);
        numCredits++;
        return;
    }

    public void setCredit (String what, String who, int num) {
        if (num >= MAX_CREDITS) {
            System.out.println ("WARNING: Overflow credits (exceed " + MAX_CREDITS + ") for code " + code);
            return;
        }

        if (num >= numCredits) {
            numCredits = num + 1;
        }

        if (credits[num] == null) {
            credits[num] = new Credit (what, who);

        } else {
            credits[num].what = what;
            credits[num].who = who;
        }

        return;
    }

    public int getNumCredits () {
        return numCredits;
    }

    public Credit getCredit (int num) {
        return credits[num];
    }

    /*
     **************************************************************************
     * Get and Set USER DATA
     **************************************************************************
     */
    public int getGrade () {
        return grade;
    }

    public int getGradeMod () {
        return gradeMod;
    }

    public double getGradeValue () {
        if (gradeValue > 0.0) {
            return gradeValue;

        } else {
            return getAutoGradeValue ();
        }
    }

    public static String valueToString (double value) {
        long dollars = (long) value;
        long cents = Math.round (100 * (value - dollars));

        if (cents == 0) {
            return "$" + dollars;

        } else if (cents < 10) {
            return "$" + dollars + "." + cents + "0";

        } else {
            return "$" + dollars + "." + cents;
        }
    }

    public String getGradeValueString () {
        return valueToString (getGradeValue());
    }

    public void setGradeValue (double newGradeValue) {
        gradeValue = newGradeValue;
    }

    public boolean isGradeValueOverride () {
        return (gradeValue > 0.0);
    }

    public boolean isGradeModUsed () {
        return (grade > GRADE_REPRINT);
    }

    public void setGrade (int newGrade) {
        grade = newGrade;

        if (grade <= GRADE_REPRINT) {
            gradeMod = GRADE_MOD_NONE;
        }
    }

    public void setGradeMod (int newGradeMod) {
        if (grade > GRADE_REPRINT) {
            gradeMod = newGradeMod;
        }
    }

    public void setGradeAndMod (String gradeStr) {
        grade = GRADE_NONE;
        gradeMod = GRADE_MOD_NONE;

        if ((gradeStr == null) || gradeStr.equals("")) {
            return;
        }

        while (gradeStr.endsWith("+")) {
            gradeStr = gradeStr.substring (0, gradeStr.length() - 1);
            gradeMod++;
        }
        while (gradeStr.endsWith("-")) {
            gradeStr = gradeStr.substring (0, gradeStr.length() - 1);
            gradeMod--;
        }

        for (int i=0; i<=GRADE_MAX; i++) {
            if (gradeStr.toUpperCase().equals(gradeAbbreviations[i].toUpperCase())) {
                grade = i;
                return;
            }
        }

        System.out.println ("Unknown Grade String: '" + gradeStr + "'");
        grade = GRADE_NONE;
        gradeMod = GRADE_MOD_NONE;
    }

    public String getGradeAndModAbbreviation () {
        String gradeStr = Issue.getAbbreviationForGrade (grade);
        int tempMod = gradeMod;

        while (tempMod > 0) {
            gradeStr = gradeStr + "+";
            tempMod--;
        }
        while (tempMod < 0) {
            gradeStr = gradeStr + "-";
            tempMod++;
        }
        return gradeStr;
    }

    public String getGradeAbbreviationOrBlank () {
        if (grade == GRADE_NONE) {
            return " ";

        } else {
            return getGradeAndModAbbreviation ();
        }
    }

    public String getComment () {
        return comment;
    }

    public void setComment (String newComment) {
        if ((comment == null) && (newComment == null)) {
            return;
        }
        if ((comment != null) && (newComment != null) && comment.equals(newComment)) {
            return;
        }
        comment = newComment;
    }

    /*
     **************************************************************************
     * STATIC METHODS
     **************************************************************************
     */
    public static String getAbbreviationForGrade (int anyGrade) {
        if ((anyGrade < 0) || (anyGrade > GRADE_MAX)) {
            return "";

        } else {
            return gradeAbbreviations[anyGrade];
        }
    }

    public static String getNameForGrade (int anyGrade) {
        if ((anyGrade < 0) || (anyGrade > GRADE_MAX)) {
            return "";

        } else {
            return gradeNames[anyGrade];
        }
    }
}
