/******************************************************************************
 * Module Name : $RCSfile: IssueRange.java,v $
 *
 * Description : Range of issues, handy for compacting and displaying in a
 *                short form.
 *
 * Updates:
 *   $Log: IssueRange.java,v $
 *   Revision 1.10  2003/01/23 06:37:46  jcouper
 *   Version 1.1 Changes
 *
 *   Revision 1.9  2002/07/13 02:22:29  jcouper
 *   job:157
 *   Still tweaking after changing grading and valuation.
 *
 *   Revision 1.8  2001/06/16 01:49:09  jcouper
 *   Commit of final release 1.0.3
 *
 *   Revision 1.7  2000/03/18 00:46:13  Jonathan
 *   Big changes in the way the IndexDialog works, including a new CardDialog.
 *   New preferences for column display.  Permit renaming of title files.
 *
 *   Revision 1.6  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.5  1999/11/28 07:13:41  Jonathan
 *   Added Flags to Titles, and Polished AllTitlesMatrix report
 *
 *   Revision 1.4  1999/11/27 09:56:58  Jonathan
 *   Add filtering placeholders.  Doesn't quite do filtering, but it's getting
 *   jolly close.
 *
 *   Revision 1.3  1999/10/24 22:37:55  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.util.*;

public class IssueRange extends Vector {
    /*
     **************************************************************************
     * Constructors.  We add a handy constructor to create an issue range
     * consisting of a single issue.
     **************************************************************************
     */
    /*
     * Single issue range.
     */
    public IssueRange (Issue newIssue) {
        super ();
        addElement (newIssue);
    }

    /*
     * Empty issue range, used when building new title definitions.
     */
    public IssueRange () {
        super ();
    }

    /*
     **************************************************************************
     * Override our inherited methods.
     **************************************************************************
     */
    public String toString () {
        if (size() <= 0) {
            return "<Empty>";

        } else if (size() == 1) {
            return getStartIssue().getCode();

        } else {
            return getStartIssue().getCode () + " .. " + getEndIssue().getCode();
        }
    }

    public void addElement (Object object) {
        if (! (object instanceof Issue)) {
            System.out.println ("IssueRange is Vector of Issue");
            System.exit (1);
        }
        super.addElement (object);
    }

    /*
     **************************************************************************
     * Get and set methods.
     **************************************************************************
     */
    public Issue getStartIssue () {
        if (size() <= 0) {
            return null;

        } else {
            return (Issue) elementAt (0);
        }
    }

    public Issue getEndIssue () {
        if (size() <= 0) {
            return null;

        } else {
            return (Issue) elementAt (size() - 1);
        }
    }

    /*
     * Get the total number of issues in this title actually held.
     */
    public int numIssues (TitleRef filterTitleRef) {
        int count = 0;

        /*
         * Deal quickly with the simple case of no filtering, and including
         * any grade.
         */
        if (filterTitleRef == null) {
            return size();
        }

        /*
         * OK, either we're interested in held issues only, and/or the title
         * reference is applying a filter.  We need to check more carefully.
         */
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            Issue issue = (Issue) e.nextElement();
            if (!filterTitleRef.includesIssue (issue)) {
                continue;
            }
            count = count + 1;
        }

        return count;
    }

    /*
     * Get the graded values of all the titles in this issue range.
     * May include a filter, which could be a held-only filter.
     */
    public double getGradeValue (TitleRef filterTitleRef) {
        double gradeValue = 0.0;
        double gradeValueTotal = 0.0;

        /*
         * OK, either we're interested in held issues only, and the title
         * reference is applying a filter.
         */
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            Issue issue = (Issue) e.nextElement();
            if ((filterTitleRef != null) && !filterTitleRef.includesIssue (issue)) {
                continue;
            }
            gradeValueTotal = gradeValueTotal + issue.getGradeValue ();
        }

        return gradeValueTotal;
    }

    /*
     * Are all these issues valued.
     * May include a filter, which could be a held-only filter.
     */
    public boolean allValued (TitleRef filterTitleRef) {
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            Issue issue = (Issue) e.nextElement();
            if ((filterTitleRef != null) && !filterTitleRef.includesIssue (issue)) {
                continue;
            }
            if (issue.getNmValue() == 0.0) {
                return false;
            }
        }

        return true;
    }

    /*
     * Number that are valued.
     * May include a filter, which could be a held-only filter.
     */
    public int numValued (TitleRef filterTitleRef) {
        int total = 0;

        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            Issue issue = (Issue) e.nextElement();
            if ((filterTitleRef != null) && !filterTitleRef.includesIssue (issue)) {
                continue;
            }
            if (issue.getNmValue() > 0.0) {
                total++;
            }
        }

        return total;
    }

    public Issue elementAt (TitleRef filterTitleRef, int pos) {
        /*
         * Deal quickly with the simple case of no filtering, and including
         * any grade.
         */
        if (filterTitleRef == null) {
            return (Issue) elementAt (pos);
        }

        /*
         * Count with filter.
         */
        int count = 0;
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            Issue issue = (Issue) e.nextElement();
            if (! filterTitleRef.includesIssue (issue)) {
                continue;
            }
            if (pos == count) {
                return issue;
            }
            count = count + 1;
        }

        return null;
    }

    /*
     **************************************************************************
     * Utility methods.
     **************************************************************************
     */
    /*
     * Does the range we contain overlap with the specified upper and lower numbers?
     */
    public boolean isOverlap (int lower, int upper) {
        if (lower > upper) {
            return false;
        }
        if (size() <= 0) {
            return false;
        }
        if ((!getStartIssue().isCodeNumeric()) || (!getEndIssue().isCodeNumeric())) {
            return false;
        }
        int startIssue = getStartIssue().getCodeAsInt();
        int endIssue = getEndIssue().getCodeAsInt();
        if (startIssue > endIssue) {
            return false;
        }
        if (lower > endIssue) {
            return false;
        }
        if (upper < startIssue) {
            return false;
        }
        return true;
    }
}