/******************************************************************************
 * Module Name : $RCSfile: Title.java,v $
 *
 * Description : Title, belonging to a card in a collection.
 *                This is just a vector of issue ranges, with a default name.
 *                Each IssueRange is a vector of issues.
 *
 * Note: The title loads the issue data from...
 *  - The template file (.tem, which contains the generic data)
 *  - The user data file (.dat, which contains the user specific info)
 * ...and from these we create ourself.
 *
 * Note: Vector of IssueRange objects.
 *
 * Updates:
 *   $Log: Title.java,v $
 *   Revision 1.27  2003/09/09 22:30:28  jcouper
 *   161: Add support for / Info of any form
 *
 *   Revision 1.26  2003/01/23 06:37:46  jcouper
 *   Version 1.1 Changes
 *
 *   Revision 1.25  2002/07/14 03:48:31  jcouper
 *   job:159
 *   Option to over-ride the automatic value.
 *
 *   Revision 1.24  2002/07/13 02:22:29  jcouper
 *   job:157
 *   Still tweaking after changing grading and valuation.
 *
 *   Revision 1.23  2002/07/13 01:06:17  jcouper
 *   job:157
 *   Add +/- for grades.  (Also fixed 147 at the same time).
 *
 *   Revision 1.22  2002/07/07 00:53:42  jcouper
 *   job:153
 *   Filenames must be lowercase, and special issues uppercase.
 *
 *   Revision 1.21  2001/10/11 21:28:04  jcouper
 *   142: Added _ALT option to Issue menu
 *
 *   Revision 1.20  2001/10/11 20:32:02  jcouper
 *   146: Added WANT_COMPLETE
 *
 *   Revision 1.19  2001/10/08 22:12:37  jcouper
 *   146: Two new flags
 *
 *   Revision 1.18  2001/10/07 03:18:38  jcouper
 *   120: Can now edit HTML text
 *
 *   Revision 1.17  2001/07/29 00:47:29  jcouper
 *   job:137
 *   Added support for credits, and re-vamped the EditDialog.
 *
 *   Revision 1.16  2001/06/16 01:49:09  jcouper
 *   Commit of final release 1.0.3
 *
 *   Revision 1.15  2001/04/04 07:34:53  jcouper
 *   Support credits, filters, new columns, etc, etc.  Stable release.
 *
 *   Revision 1.14  2000/08/24 14:27:56  jcouper
 *   Support /html lines.
 *
 *   Revision 1.13  2000/03/18 00:46:13  Jonathan
 *   Big changes in the way the IndexDialog works, including a new CardDialog.
 *   New preferences for column display.  Permit renaming of title files.
 *
 *   Revision 1.12  2000/02/18 22:35:35  Jonathan
 *   Add basic application settings support.
 *
 *   Revision 1.11  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.10  1999/11/28 07:13:41  Jonathan
 *   Added Flags to Titles, and Polished AllTitlesMatrix report
 *
 *   Revision 1.9  1999/11/27 09:56:58  Jonathan
 *   Add filtering placeholders.  Doesn't quite do filtering, but it's getting
 *   jolly close.
 *
 *   Revision 1.8  1999/11/15 22:26:59  Jonathan
 *   Report stuff compiles at least.  Doesn't run, mind you.
 *
 *   Revision 1.7  1999/10/25 08:33:07  Jonathan
 *   v0_9_2
 *
 *   Revision 1.6  1999/10/24 22:37:55  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.io.*;
import java.util.*;
import standup.util.*;
import standup.*;

public class Title extends Vector {
    /*
     * Static variables.
     */
    public static final String flagInfoIsCode = "INFO_IS_CODE";
    public static final String templateFlags[] = {
            flagInfoIsCode};

    public static final String flagWantComplete = "WANT_COMPLETE";
    public static final String flagWantNone = "WANT_NONE";
    public static final String userFlags[] = {
            flagWantComplete,
            flagWantNone};

    String keyname = null;
    String renameKeyname = null;            // Rename, waiting for save of index
    String defaultTitleName = null;
    String htmlDescription = "";            // Optional HTML title description
    File templateFile = null;
    File userFile = null;
    Vector templateFlagsSet = new Vector();
    Vector userFlagsSet = new Vector();

    boolean templateReadPending = true;     // Have we tried to load?

    protected boolean userDataModified = false;
    protected boolean templateDataModified = false;

    /*
     **************************************************************************
     * Constructor.  Sets things up, but doesn't load data until requested.
     **************************************************************************
     */
    public Title (String newKeyname, File newTemplateFile, File newUserFile) {
        keyname = newKeyname;

        templateFile = newTemplateFile;
        userFile = newUserFile;

        templateReadPending = templateFile.exists();
    }

    /*
     **************************************************************************
     * Override our inherited methods.
     **************************************************************************
     */
    /*
     * Note that size() returns the number of IssueRanges, not the number of
     * issues.  Use numIssues() to get the number of issues.
     */
    public int size () {
        loadDataNow();
        return super.size();
    }

    public String toString () {
        return getAnyName ();
    }

    public void addElement (Object object) {
        loadDataNow();
        if (! (object instanceof IssueRange)) {
            System.out.println ("Title is Vector of IssueRange");
            System.exit (1);
        }
        super.addElement (object);
        templateDataModified = true;
    }

    public void setElementAt (Object object, int pos) {
        loadDataNow();
        if (! (object instanceof IssueRange)) {
            System.out.println ("Title is Vector of IssueRange");
            System.exit (1);
        }
        super.setElementAt (object, pos);
        templateDataModified = true;
    }

    public void insertElementAt (Object object, int pos) {
        loadDataNow();
        if (! (object instanceof IssueRange)) {
            System.out.println ("Title is Vector of IssueRange");
            System.exit (1);
        }
        super.insertElementAt (object, pos);
        templateDataModified = true;
    }

    public void removeElementAt (int pos) {
        super.removeElementAt (pos);
        templateDataModified = true;
    }

    /*
     **************************************************************************
     * Get and set methods.
     **************************************************************************
     */
    public String getKeyname () {
        return keyname;
    }

    /*
     * Return our best guess at a name for this title.
     */
    public String getDefaultTitleName () {
        loadDataNow();
        return defaultTitleName;
    }

    public String getAnyName () {
        loadDataNow();
        if ((defaultTitleName != null) && !defaultTitleName.equals("")) {
            return defaultTitleName;
        }
        return keyname;
    }

    public void setDefaultTitleName (String newDefaultTitleName) {
        loadDataNow();
        if ((defaultTitleName == null) && (newDefaultTitleName == null)) {
            return;
        }
        if ((defaultTitleName != null) && (newDefaultTitleName != null) && defaultTitleName.equals(newDefaultTitleName)) {
            return;
        }
        defaultTitleName = newDefaultTitleName;
        templateDataModified = true;
    }

    /*
     * The relates to our option flags on the title, not the modified flags.
     * Bit confusing there on the old naming front.  Sorry.
     */
    public boolean hasFlag (String flagName) {
        loadDataNow();
        return (templateFlagsSet.contains (flagName) || userFlagsSet.contains (flagName));
    }
    public void setFlag (String flagName) {
        loadDataNow();
        if (hasFlag (flagName)) {
            return;
        }
        for (int i=0; i<Title.templateFlags.length; i++) {
            if (templateFlags[i].equals (flagName)) {
                templateFlagsSet.addElement (flagName);
                templateDataModified = true;
                return;
            }
        }
        for (int i=0; i<Title.userFlags.length; i++) {
            if (userFlags[i].equals (flagName)) {
                userFlagsSet.addElement (flagName);
                userDataModified = true;
                return;
            }
        }
        return;
    }
    public void clearFlag (String flagName) {
        loadDataNow();
        if (!hasFlag (flagName)) {
            return;
        }
        for (int i=0; i<Title.templateFlags.length; i++) {
            if (templateFlags[i].equals (flagName)) {
                templateFlagsSet.removeElement (flagName);
                templateDataModified = true;
                return;
            }
        }
        for (int i=0; i<Title.userFlags.length; i++) {
            if (userFlags[i].equals (flagName)) {
                userFlagsSet.removeElement (flagName);
                userDataModified = true;
                return;
            }
        }
        return;
    }

    public String getHtmlDescription () {
        loadDataNow();
        return htmlDescription;
    }
    public void setHtmlDescription (String value) {
        loadDataNow();
        htmlDescription = value;
        templateDataModified = true;
    }

    /*
     **************************************************************************
     * Utility routines.
     **************************************************************************
     */
    /*
     * Find an issue by number.  Brute force lookup.  Not pretty.  Note
     * also that we don't force codes to be unique.
     */
    public Issue getIssueByCode (String issueCodeToMatch) {
        loadDataNow();
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();

            for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                Issue issue = (Issue) e2.nextElement();
                if (issue.getCode().equals(issueCodeToMatch)) {
                    return issue;
                }
            }
        }
        return null;
    }

    /*
     * Find an issue by place in the overall sequence of issues in
     * this title.  issueOffset is zero index.
     */
    public Issue getIssueByArrayPos (int issueOffset) {
        return getIssueByArrayPos (issueOffset, null);
    }
    public Issue getIssueByArrayPos (int issueOffset, TitleRef filterTitleRef) {
        int counted = 0;
        int rangeSize = 0;

        loadDataNow();
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            rangeSize = issueRange.numIssues (filterTitleRef);

            if ((issueOffset - counted) < rangeSize) {
                return (Issue) issueRange.elementAt (filterTitleRef, issueOffset - counted);
            }
            counted = counted + rangeSize;
        }
        return null;
    }

    /*
     * Finds number of consecutive grades from pos.  Must be in same
     * issue range.  Don't support filtering.
     */
    public int consecutiveGradeFromPos (int issueOffset) {
        int counted = 0;
        int rangeSize = 0;

        loadDataNow();
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            rangeSize = issueRange.size();

            if ((issueOffset - counted) < rangeSize) {
                int consecutive = 1;
                int rangePos = issueOffset - counted;
                Issue issue = (Issue) issueRange.elementAt (rangePos);
                int firstGrade = issue.getGrade ();

                rangePos++;
                while (rangePos < rangeSize) {
                    issue = (Issue) issueRange.elementAt (rangePos);

                    if (firstGrade != issue.getGrade()) {
                        break;
                    }
                    consecutive++;
                    rangePos++;
                }
                return consecutive;
            }
            counted = counted + rangeSize;
        }
        return 0;
    }

    /*
     * Step through the issues and derive from our mark dates.
     */
    public void computeDerivedDates () {
        Issue precedingIssue = null;

        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();

            for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                Issue issue = (Issue) e2.nextElement();

                // Note that this won't trample a mark date.
                if (precedingIssue != null) {
                    issue.setCoverDateFromPrevious (precedingIssue);
                }
                precedingIssue = issue;
            }
        }
    }

    /*
     * Return true iff all issues have a known date.
     */
    public boolean allIssuesDated () {
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();

            for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                Issue issue = (Issue) e2.nextElement();

                CoverDate date = issue.getCoverDate();

                if (!date.isDefined()) {
                    return false;
                }
            }
        }
        return true;
    }

    /*
     * Return true iff all issues have a non-zero NM value.
     */
    public boolean allIssuesValued () {
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();

            for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                Issue issue = (Issue) e2.nextElement();

                double value = issue.getNmValue();

                if (value == 0.0) {
                    return false;
                }
            }
        }
        return true;
    }

    /*
     * Return a comma separated list of issues held, optionally
     * with grades, optionally with values.
     */
    protected String runEntry (Issue runStart,
                               Issue runEnd,
                               boolean values,
                               boolean grades) {


        int num;
        String result = "";

        if (runStart == runEnd) {
            num = 1;
            result = runStart.getCode();

        } else {
            num = 1 + (runEnd.getCodeAsInt() - runStart.getCodeAsInt());
            result = runStart.getCode() + "-" + runEnd.getCode();
        }

        if (grades || values) {
            result = result + " (";
            if (grades) {
                result = result + runStart.getGradeAndModAbbreviation();
            }
            if (values) {
                if (grades) {
                    result = result + "  ";
                }
                result = result + runStart.getGradeValueString();
            }

            result = result + ")";
        }

        return result;
    }

    public String getRunList (boolean values, boolean grades) {
        StringBuffer list = new StringBuffer ();

        Issue runStart = null;
        Issue runEnd = null;

        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();

            for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                Issue issue = (Issue) e2.nextElement();

                // Not held?
                if (! issue.isHeld()) {
                    continue;
                }

                // In a run, extend?
                if ((runStart != null) &&
                    (issue.isCodeNumeric ()) &&
                    (issue.getCodeAsInt() == (runEnd.getCodeAsInt() + 1)) &&
                    (issue.getGrade() == runStart.getGrade()) &&
                    (issue.getGradeMod() == runStart.getGradeMod()) &&
                    (issue.getGradeValue() == runStart.getGradeValue())) {

                    runEnd = issue;
                    continue;
                }

                // Finish any run
                if (runStart != null) {
                    if (list.length() > 0) { list.append (", "); }
                    list.append (runEntry (runStart, runEnd, values, grades));
                    runStart = null;
                    runEnd = null;
                }

                // Start a run?
                if (issue.isCodeNumeric ()) {
                    runStart = issue;
                    runEnd = issue;

                // One-off issue.
                } else {
                    if (list.length() > 0) { list.append (", "); }
                    list.append (runEntry (issue, issue, values, grades));
                }
            }
        }

        // Incomplete run?
        if (runStart != null) {
            if (list.length() > 0) { list.append (", "); }
            list.append (runEntry (runStart, runEnd, values, grades));
        }

        return list.toString ();
    }

    /*
     **************************************************************************
     * MODIFIED FLAG METHODS.
     **************************************************************************
    /*
     * Have we been modified since last load or save?  I.e. should we be
     * presented to the user in a Save list.
     */
    public boolean isModified () {
        return (templateDataModified || userDataModified);
    }

    public void flagUserDataModified () {
        userDataModified = true;
    }

    public void flagTemplateDataModified () {
        templateDataModified = true;
    }

    /*
     **************************************************************************
     * DATA LOAD METHODS
     **************************************************************************
    /*
     * Get the total number of issues in this title.  The filterTitleRef may
     * be null, which implies that there is no filtering required, and we can
     * skip the filtering callback.  Or, it may be an explicit filter, and/or
     * a held-only filter.
     */
    public int numIssues () {
        return numIssues (null);
    }
    public int numIssues (TitleRef filterTitleRef) {
        int count = 0;

        loadDataNow();
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            count = count + issueRange.numIssues (filterTitleRef);
        }

        return count;
    }

    public double getGradeValue () {
        return getGradeValue (null);
    }
    public double getGradeValue (TitleRef filterTitleRef) {
        double gradeValue = 0.0;
        double gradeValueTotal = 0.0;

        loadDataNow();
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            gradeValue = issueRange.getGradeValue (filterTitleRef);
            gradeValueTotal = gradeValueTotal + gradeValue;
        }

        return gradeValueTotal;
    }

    public boolean allValued () {
        return allValued (null);
    }
    public boolean allValued (TitleRef filterTitleRef) {
        loadDataNow();

        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            if (! issueRange.allValued (filterTitleRef)) {
                return false;
            }
        }
        return true;
    }

    public int numValued () {
        return numValued (null);
    }
    public int numValued (TitleRef filterTitleRef) {
        loadDataNow();

        int total = 0;
        for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
            IssueRange issueRange = (IssueRange) e.nextElement();
            total = total + issueRange.numValued (filterTitleRef);
        }
        return total;
    }

    /*
     * This only works if the last issue of the title is numeric.  We
     * increment by one and pop another on the end.  We return true
     * iff we added another issue.  Note, we don't clone the credits.
     */
    public boolean cloneLastIssue () {
        if (size() <= 0) {
            return false;
        }

        // Get the last range, and ensure it is numeric.
        IssueRange lastRange = (IssueRange) elementAt (size() - 1);
        Issue lastIssue = lastRange.getEndIssue();
        if (!lastIssue.isCodeNumeric()) {
            return false;
        }

        // Copy/Extend the parameters and make a new issue.
        Issue newIssue = new Issue (
            String.valueOf (lastIssue.getCodeAsInt() + 1),
            lastIssue.getNmValue(),
            lastIssue.getGrade(),
            lastIssue.getGradeMod(),
            null,
            null);

        // Add the new issue.
        lastRange.addElement (newIssue);

        // Derive new dates.
        newIssue.setCoverDateFromPrevious (lastIssue);

        // We're modified.
        templateDataModified = true;

        return true;
    }


    /*
     * This makes an _ALT issue of a numeric issue, which will probably
     * involve splitting a range.  This is YUCKY!
     */
    public boolean makeAlternateIssue (Issue issue) {
        int idx = 0;
        IssueRange baseRange = null;

        // Find the issue range containing this one.
        for (idx=0 ; idx<size() ; idx++) {
            baseRange = (IssueRange) elementAt (idx);
            if (baseRange.contains (issue)) {
                break;
            }
        }
        if (idx >= size()) {
            System.out.println ("Issue not in Title!");
            System.exit (1);
        }

        // Where is it in that range?
        int pos = baseRange.indexOf (issue);

        // Copy/Extend the parameters and make a new issue.
        Issue altIssue = new Issue (
            issue.getCode() + "_ALT",
            issue.getNmValue(),
            Issue.GRADE_NONE,
            Issue.GRADE_MOD_NONE,
            issue.getInfo() + " (Alternate Cover)",
            null);

        altIssue.copyCoverDate (issue);

        // Add the new range.
        IssueRange altRange = new IssueRange(altIssue);
        insertElementAt (altRange, idx + 1);

        // Do we need to split the original range?
        if (pos < (baseRange.size() - 1)) {
            IssueRange topRange = new IssueRange();
            for (int i=baseRange.size()-1; i>pos; i--) {
                // Take from top of base and jump to at base of top
                topRange.insertElementAt (baseRange.elementAt(i), 0);
                baseRange.removeElementAt (i);
            }

            insertElementAt (topRange, idx + 2);
        }

        // We're modified.
        templateDataModified = true;

        return true;
    }


    /*
     * Append a new issue to the last of our internal list of issue ranges.  If
     * it extends the range, then that's all well and good.  If it doesn't then
     * we start a new IssueRange.
     *
     * Date parameters are the parameters from a directly preceeding /Date command,
     * or null if none.
     */
    protected void appendIssue (Issue newIssue, String dateParameters) {
        boolean extendLastRange = false;

        /*
         * The only time we will extend the existing range is if the new issue
         * has a numeric code which is 1 + the numeric code of the last issue
         * of the last range in our existing list.
         */
        if (size () > 0) {
            IssueRange lastRange = (IssueRange) elementAt (size() - 1);
            Issue lastIssue = lastRange.getEndIssue();

            Integer newCodeAsInteger = newIssue.getCodeAsInteger ();
            Integer lastCodeAsInteger = lastIssue.getCodeAsInteger ();

            if ((newCodeAsInteger != null) && (lastCodeAsInteger != null)) {
                int newCode = newCodeAsInteger.intValue();
                int lastCode = lastCodeAsInteger.intValue();

                if (newCode == (lastCode + 1)) {
                    extendLastRange = true;
                }
            }
        }

        if (extendLastRange) {
            IssueRange lastRange = (IssueRange) elementAt (size() - 1);
            lastRange.addElement (newIssue);

        } else {
            addElement (new IssueRange (newIssue));
        }

        /*
         * Now we set the date.  We only do explicit dates for now.  The
         * derived dates, we can do when we've finished loading the file/
         */
        if (dateParameters != null) {
            newIssue.setCoverDateExplicit (dateParameters);
        }

        templateDataModified = true;
    }

    /*
     * Attempt to load data from our files.
     */
    protected void loadDataNow () {

        // Avoid double-reading in some sequences
        if (!templateReadPending) {
            return;
        }
        templateReadPending = false;

        // Did we do a backwards compatability template/user tweak, meaning
        // we need to do a rewrite?
        boolean templateTweak = false;
        boolean userTweak = false;

        /*
         * First, open the template file and parse for entries.
         */
        try {
            BufferedReader input = new BufferedReader (new FileReader (templateFile));
            String line = null;
            String dateParameters = null;
            Issue lastIssue = null;

            while ((line = input.readLine()) != null) {

                // Trim lines which start with #
                if ((line.length() == 0) || (line.charAt(0) == '#')) {
                    continue;
                }

                // Get first token.
                MyStringTokenizer tokenizer = new MyStringTokenizer (line.trim());
                String firstToken = null;

                try {
                    firstToken = tokenizer.nextToken();
                }
                catch (NoSuchElementException nse) {
                    continue;
                }

                /*
                 * /[Nn]ame <name>
                 * Sets the default name.
                 */
                if (firstToken.toLowerCase().equals("/name")) {
                    defaultTitleName = tokenizer.remainder().trim();
                    if (defaultTitleName.equals("")) {
                        System.out.println ("Missing Name in '" + templateFile + "'");
                        System.out.println ("> " + line);
                        continue;
                    }

                /*
                 * /[Dd]ate [<start-YYYYMM> [step-indicator]]
                 *
                 * Start or stops cover date sequencing.  Keep the parameters safe, they
                 * apply to the next issue.
                 */
                } else if (firstToken.toLowerCase().equals("/date")) {
                    dateParameters = tokenizer.remainder().trim();

                /*
                 * /[Ff]lags [<any-flag>...]
                 *
                 * Arbitary flag tokens.
                 */
                } else if (firstToken.toLowerCase().equals("/flags")) {
                    while (tokenizer.hasMoreTokens()) {
                        templateFlagsSet.addElement ((String) tokenizer.nextToken());
                    }

                /*
                 * /[Hh]tml [Blurb...]
                 * /[Hh]tml [Blurb...]
                 *
                 * Blurb.  Note, we turn empty lines to space lines, otherwise
                 * we lose them when we write them out, since the tokenizer
                 * discards them.
                 */
                } else if (firstToken.toLowerCase ().equals ("/html")) {
                    String fragment = tokenizer.remainder ().trim ();
                    if (fragment.equals ("")) {
                        htmlDescription = htmlDescription + " \n";

                    } else {
                        htmlDescription = htmlDescription + fragment + "\n";
                    }

                /*
                 * /[Cc]redit [Who=What]
                 * /[Cc]redit [Who=What]
                 *
                 * Credits, who did what for the last issue.
                 */
                } else if (firstToken.toLowerCase ().equals ("/credit")) {
                    if (lastIssue == null) {
                        System.out.println ("Credit before Issues in '" + templateFile + "'");
                        System.out.println ("> " + line);

                    } else {
                        String details = tokenizer.remainder ().trim ();
                        if (details.indexOf("=") > -1) {
                            String what = details.substring (0, details.indexOf("="));
                            String who = details.substring (details.indexOf("=") + 1);
                            lastIssue.addCredit (what, who);

                        } else {
                            System.out.println ("Bad Credit Details in '" + templateFile + "'");
                            System.out.println ("> " + line);
                        }
                    }

                /*
                 * / <extra>
                 *
                 * Extra free-form stuff, used by other programs and systems, we don't
		 * interpret.
                 */
                } else if (firstToken.toLowerCase ().equals ("/")) {
		    String extra = tokenizer.remainder ().trim ();
		    lastIssue.addExtra (extra);
		
                /*
                 * /<Bad-keyword>
                 * Spit the dummy.
                 */
                } else if (firstToken.startsWith("/")) {
                    System.out.println ("Bad Keyword in '" + templateFile + "'");
                    System.out.println ("> " + line);

                /*
                 * Else it's an issue!
                 * See help files for format.
                 */
                } else {
                    double nmValue = 0.0;
                    String infoString = null;

                    if (tokenizer.hasMoreTokens()) {
                        String nmValueString = tokenizer.nextToken();

                        nmValueString = nmValueString.substring (1);
                        try {
                            nmValue = Double.parseDouble (nmValueString);
                        }
                        catch (NumberFormatException nfe) {
                            System.out.println ("Bad Value in '" + templateFile + "'");
                            System.out.println ("> " + line);
                        }
                    }
                    if (tokenizer.hasMoreTokens()) {
                        infoString = tokenizer.remainder().trim();
                    }

                    /*
                     * Range of issues?
                     */
                    if (firstToken.indexOf("..") > 0) {
                        String startIssue = firstToken.substring(0, firstToken.indexOf(".."));
                        String endIssue = firstToken.substring(firstToken.indexOf("..") + 2);

                        try {
                            int startIssueNum = (new Integer (startIssue)).intValue();
                            int endIssueNum = (new Integer (endIssue)).intValue();
                            for (int issueNum = startIssueNum; issueNum <= endIssueNum; issueNum++) {
                                Issue newIssue = new Issue (String.valueOf(issueNum), nmValue, infoString);
                                appendIssue (newIssue, dateParameters);
                                dateParameters = null;
                                lastIssue = newIssue;
                            }
                        }
                        catch (NumberFormatException nfe) {
                            System.out.println ("Non-Numeric Issue Range in '" + templateFile + "'");
                            System.out.println ("> " + line);
                            continue;
                        }

                    /*
                     * Single issue case.
                     */
                    } else {
                        Issue newIssue = new Issue (firstToken, nmValue, infoString);
                        appendIssue (newIssue, dateParameters);
                        dateParameters = null;
                        lastIssue = newIssue;
                    }
                }
            }
            input.close();
        }
        catch (IOException io) {
        }

        /*
         * Now compute our derived dates, since the template is all loaded.
         */
        computeDerivedDates ();

        /*
         * Then check for user data.
         */
        try {
            BufferedReader input = new BufferedReader (new FileReader (userFile));
            String line = null;
            Issue lastIssue = null;

            while ((line = input.readLine()) != null) {

                // Trim lines which start with #
                if ((line.length() == 0) || (line.charAt(0) == '#')) {
                    continue;
                }

                // Get first token.
                MyStringTokenizer tokenizer = new MyStringTokenizer (line.trim());
                String firstToken = null;

                try {
                    firstToken = tokenizer.nextToken();
                }
                catch (NoSuchElementException nse) {
                    continue;
                }

                /*
                 * /[Pp]referred[Gg]rade <grade-code>
                 * Deprecated
                 */
                if (firstToken.toLowerCase().equals("/preferredgrade")) {

                /*
                 * /[Ff]lags [<any-flag>...]
                 *
                 * Arbitary flag tokens.
                 */
                } else if (firstToken.toLowerCase().equals("/flags")) {
                    while (tokenizer.hasMoreTokens()) {
                        userFlagsSet.addElement ((String) tokenizer.nextToken());
                    }

                /*
                 * /[Vv]alue $<value>
                 *
                 * Explicit value for this held grade.
                 */
                } else if (firstToken.toLowerCase().equals("/value")) {

                    String gradeValueString = tokenizer.nextToken();

                    gradeValueString = gradeValueString.substring (1);
                    try {
                        double gradeValue = Double.parseDouble (gradeValueString);

                        if (lastIssue == null) {
                            System.out.println ("Value before Issues in '" + userFile + "'");
                            System.out.println ("> " + line);

                        } else {
                            lastIssue.setGradeValue (gradeValue);
                        }

                    }
                    catch (NumberFormatException nfe) {
                        System.out.println ("Bad Value in '" + userFile + "'");
                        System.out.println ("> " + line);
                    }

                /*
                 * /<Bad-keyword>
                 * Spit the dummy.
                 */
                } else if (firstToken.startsWith("/")) {
                    System.out.println ("Bad Keyword in '" + userFile + "'");
                    System.out.println ("> " + line);

                /*
                 * Else it's an issue!
                 * See help files for format.
                 */
                } else {
                    String commentString = null;
                    String gradeString = null;

                    if (tokenizer.hasMoreTokens()) {
                        gradeString = tokenizer.nextToken();

                        // Backwards compatability for pre 1.0.6 releases.
                        if (gradeString.equals ("F/VF")) { userTweak = true; gradeString = "FN++"; }
                        if (gradeString.equals ("M/NM")) { userTweak = true; gradeString = "NM "; }
                        if (gradeString.equals ("?")) { userTweak = true; gradeString = "<???>"; }
                        if (gradeString.equals ("ORD")) { userTweak = true; gradeString = "<ord>"; }
                        if (gradeString.equals ("R")) { userTweak = true; gradeString = "<rep>"; }
                    }

                    if (tokenizer.hasMoreTokens()) {
                        commentString = tokenizer.remainder().trim();
                    }

                    if (firstToken.indexOf("..") > 0) {
                        String startIssue = firstToken.substring(0, firstToken.indexOf(".."));
                        String endIssue = firstToken.substring(firstToken.indexOf("..") + 2);

                        try {
                            int startIssueNum = (new Integer (startIssue)).intValue();
                            int endIssueNum = (new Integer (endIssue)).intValue();
                            for (int issueNum = startIssueNum; issueNum <= endIssueNum; issueNum++) {
                                Issue issue = this.getIssueByCode (String.valueOf(issueNum));
                                if (issue == null) {
                                    System.out.println ("Unknown Issue in '" + userFile + "'");
                                    System.out.println ("> " + line);

                                } else {
                                    issue.initUserData (gradeString, commentString);
                                    lastIssue = issue;
                                }
                            }
                        }
                        catch (NumberFormatException nfe) {
                            System.out.println ("Non-Numeric Issue Range in '" + userFile + "'");
                            System.out.println ("> " + line);
                            continue;
                        }

                    } else {
                        Issue issue = this.getIssueByCode (firstToken);
                        if (issue == null) {
                            System.out.println ("Unknown Issue in '" + userFile + "'");
                            System.out.println ("> " + line);

                        } else {
                            issue.initUserData (gradeString, commentString);
                            lastIssue = issue;
                        }
                    }
                }
            }
            input.close();
        }
        catch (IOException io) {
            // No problem, we don't require a user data file.
        }

        // Did we tweak data while reading it?
        userDataModified = userTweak;
        templateDataModified = templateTweak;
    }

    /*
     **************************************************************************
     * DATA SAVE METHODS
     **************************************************************************
     */
    /*
     * Save whatever needs saving.  This is the only public save method.
     */
    public void save () {
        if (templateDataModified) {
            saveTemplateData ();
            saveUserData ();

        } else if (userDataModified) {
            saveUserData ();
        }
    }

    /*
     * Mark file file for rename.  We don't do the rename until we save the
     * index file.
     */
    public void markForRename (String newKeyname) {
        if (keyname.equals (newKeyname)) {
            renameKeyname = null;

        } else {
            renameKeyname = newKeyname;
        }
    }

    /*
     * Mark file file for rename.  We don't do the rename until we save the
     * index file.
     */
    public boolean isRenameRequired () {
        return (renameKeyname != null);
    }

    /*
     * Rename the file, if we have done a rename.  This is called when we
     * save the index file.
     */
    public void renameIfRequired () {
        if (renameKeyname == null) {
            return;
        }
        File renamedTemplateFile = new File (templateFile.getParent(), renameKeyname + ".tem");
        File renamedUserFile = new File (userFile.getParent(), renameKeyname + ".dat");

        templateFile.renameTo (renamedTemplateFile);
        templateFile = renamedTemplateFile;

        userFile.renameTo (renamedUserFile);
        userFile = renamedUserFile;

        keyname = renameKeyname;
        renameKeyname = null;
    }

    /*
     * Save user data.  I think we can safely assume it's been loaded, but it
     * doesn't hurt to make sure!
     */
    protected void saveUserData () {
        loadDataNow();
        try {
            BufferedWriter output = new BufferedWriter (new FileWriter (userFile));

            output.write ("#");
            output.newLine ();
            output.write ("# This file " + userFile + " generated automatically");
            output.newLine ();
            output.write ("# by Standup Comic Management Software (Version " + Application.getVersion() + ")");
            output.newLine ();
            output.write ("#");
            output.newLine ();

            if ((userFlagsSet != null) && userFlagsSet.size() > 0) {
                String line = "/Flags";

                for (Enumeration e = userFlagsSet.elements() ; e.hasMoreElements() ;) {
                    line = line + " " + ((String) e.nextElement());
                }
                output.write (line);
                output.newLine ();
            }

            for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
                IssueRange issueRange = (IssueRange) e.nextElement();

                for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                    Issue issue = (Issue) e2.nextElement();

                    String gradeCode = issue.getGradeAndModAbbreviation();
                    String comment = issue.getComment();

                    if (!gradeCode.equals("<none>")) {
                        String line = issue.getCode() + " " + gradeCode;
                        if ((issue.getComment () != null) && !issue.getComment().equals("")) {
                            line = line + " " + comment;
                        }
                        output.write (line);
                        output.newLine ();
                    }

                    if (issue.isGradeValueOverride()) {
                        String gradeValue = "$" + issue.getGradeValue();
                        String line = "/Value " + gradeValue;

                        output.write (line);
                        output.newLine ();
                    }
                }
            }
            output.write ("#");
            output.newLine ();
            output.write ("# End of User Data");
            output.newLine ();
            output.write ("#");
            output.newLine ();

            output.close();
        }
        catch (IOException io) {
        }
        userDataModified = false;
    }

    /*
     * Save template data.  I think we can safely assume it's been loaded, but it
     * doesn't hurt to make sure!
     */
    protected void saveTemplateData () {
        loadDataNow();
        try {
            BufferedWriter output = new BufferedWriter (new FileWriter (templateFile));

            output.write ("#");
            output.newLine ();
            output.write ("# This file " + templateFile + " generated automatically");
            output.newLine ();
            output.write ("# by Standup Comic Management Software (Version " + Application.getVersion() + ")");
            output.newLine ();
            output.write ("#");
            output.newLine ();

            if ((defaultTitleName != null) && !defaultTitleName.equals("")) {
                output.write ("/Name " + defaultTitleName);
                output.newLine ();
            }

            if ((templateFlagsSet != null) && templateFlagsSet.size() > 0) {
                String line = "/Flags";

                for (Enumeration e = templateFlagsSet.elements() ; e.hasMoreElements() ;) {
                    line = line + " " + ((String) e.nextElement());
                }
                output.write (line);
                output.newLine ();
            }

            if (! htmlDescription.equals ("")) {
                MyStringTokenizer tokenizer = new MyStringTokenizer (htmlDescription, "\n");
                while (tokenizer.hasMoreTokens ()) {
                    output.write ("/Html " + tokenizer.nextToken ());
                    output.newLine ();
                }
            }

	    boolean firstIssue = true;
            for (Enumeration e = this.elements() ; e.hasMoreElements() ;) {
                IssueRange issueRange = (IssueRange) e.nextElement();

                for (Enumeration e2 = issueRange.elements() ; e2.hasMoreElements() ;) {
                    Issue issue = (Issue) e2.nextElement();

                    /*
                     * Use our cover date routines to tell us if this issue date is a natural
                     * step from the preceding issue date.
                     */
                    String dateParameters = issue.getDateParameters ();
                    if (dateParameters != null) {
		        if (! firstIssue) {
                            output.write ("\n");
			}
                        if (dateParameters.equals("")) {
                            output.write ("/Date");

                        } else {
                            output.write ("/Date " + dateParameters);
                        }
                        output.newLine ();
                    }

                    String nmValue = "$" + issue.getNmValue();
                    String info = issue.getInfo();

                    String line = issue.getCode();
                    if ((info != null) && !info.equals("")) {
                        line = line + " " + nmValue + " " + info;

                    } else if (issue.getNmValue() > 0.0) {
                        line = line + " " + nmValue;
                    }
                    output.write (line);
                    output.newLine ();

                    /*
                     * Now the credits for this issue,
                     */
                    int numCredits = issue.getNumCredits ();
                    for (int i = 0; i < numCredits; i++) {
                        Credit credit = issue.getCredit (i);
                        if (credit != null) {
                            credit.what = credit.what.trim();
                            credit.who = credit.who.trim();

                            if ((! credit.who.equals("")) && (! credit.what.equals(""))) {
                                output.write ("/Credit " + credit.what + "=" + credit.who);
                                output.newLine ();
                            }
                        }
                    }

                    /*
                     * And any extra info for the issue.
                     */
		    String extra = issue.getExtra ();
	            if (! extra.equals ("")) {
        	        MyStringTokenizer tokenizer = new MyStringTokenizer (extra, "\n");
                	while (tokenizer.hasMoreTokens ()) {
                    	    output.write ("/ " + tokenizer.nextToken ());
                    	    output.newLine ();
                	}
            	    }
		    		    
		    firstIssue = false;
                }
            }
            output.write ("#");
            output.newLine ();
            output.write ("# End of User Data");
            output.newLine ();
            output.write ("#");
            output.newLine ();

            output.close();
        }
        catch (IOException io) {
        }
        templateDataModified = false;
    }
}
