/******************************************************************************
 * Module Name : $RCSfile: TitleRef.java,v $
 *
 * Description : Title reference.  This is what a card contains.  It
 *                has a reference to a canonical title object.
 *
 * Note: Why do we have TitleRef's, not just Titles?
 * a) A single title can be referred to from more than one card.
 * b) A card may refer to a subset of all the issues of a title.
 *
 * Updates:
 *   $Log: TitleRef.java,v $
 *   Revision 1.9  2003/01/23 06:37:46  jcouper
 *   Version 1.1 Changes
 *
 *   Revision 1.8  2001/04/11 08:22:24  jcouper
 *   Add Filter Support
 *
 *   Revision 1.7  2001/04/04 07:34:53  jcouper
 *   Support credits, filters, new columns, etc, etc.  Stable release.
 *
 *   Revision 1.6  2000/03/18 00:46:13  Jonathan
 *   Big changes in the way the IndexDialog works, including a new CardDialog.
 *   New preferences for column display.  Permit renaming of title files.
 *
 *   Revision 1.5  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.4  1999/11/27 09:56:59  Jonathan
 *   Add filtering placeholders.  Doesn't quite do filtering, but it's getting
 *   jolly close.
 *
 *   Revision 1.3  1999/10/24 22:37:55  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.parse;

import java.io.*;
import java.util.*;

public class TitleRef {
    // Reference to the full title info.  Private.
    private Title title = null;
    private Index index = null;

    // Name of this reference.
    String name = null;

    // Filter parmeters
    String filterString = null;
    boolean heldOnly = false;

    /*
     **************************************************************************
     * Constructor.  Sets things up, but doesn't load data until requested.
     **************************************************************************
     */
    /*
     * Unnamed title reference.
     */
    public TitleRef (Index newIndex, Title newTitle) {
        index = newIndex;
        title = newTitle;
    }

    /*
     * Named title reference.
     */
    public TitleRef (Index newIndex, Title newTitle, String newRefName) {
        index = newIndex;
        title = newTitle;
        name = newRefName;
    }

    /*
     **************************************************************************
     * Override of standard methods.
     **************************************************************************
     */
    public String toString () {
        return getAnyName();
    }

    /*
     **************************************************************************
     * Simple Get and Set type methods.
     **************************************************************************
     */
    public String getName () {
        return name;
    }

    public String getAnyName () {
        if ((name != null) && !name.equals("")) {
            return name;
        }
        return title.getAnyName();
    }

    public Title getTitle () {
        return title;
    }

    public Index getIndex () {
        return index;
    }

    /*
     **************************************************************************
     * Filter methods.
     **************************************************************************
     */
    /*
     * Get and set our individual filter parameters.
     */
    public void setFilterFromString (String newFilterString) {
        filterString = newFilterString.trim();
    }

    public String getFilterAsString () {
        return (filterString == null) ? "" : filterString;
    }

    public void setFilterHeldOnly (boolean newHeldOnly) {
        heldOnly = newHeldOnly;
    }

    public boolean getFilterHeldOnly () {
        return heldOnly;
    }

    /*
     * Quick check if we do have a filter or not.
     */
    public boolean isFiltered () {
        return heldOnly || ((filterString != null) && !filterString.equals(""));
    }

    /*
     * THE routine which determines if an issue is filtered or not.
     */
    public boolean includesIssue (Issue issue) {
        if (heldOnly) {
            if (!issue.isHeld ()) {
                return false;
            }
        }
        if ((filterString != null) && !filterString.equals("")) {
            StringTokenizer tokenizer = new StringTokenizer (filterString, ",");
            while (tokenizer.hasMoreTokens()) {
                try {
                    String token = tokenizer.nextToken ();
                    if (issue.getCode().equalsIgnoreCase (token)) {
                        return true;
                    }
                }
                catch (NoSuchElementException nse) {}
            }
            return false;
        }
        return true;
    }

    /*
     * Return the number of issues.  If there is no filtering, then
     * we have the trivial case.  With filtering, we need to do more
     * work, and get the title to filter on us.  Similiarly for
     * issue by array pos, and grade value totals.
     */
    public int numIssues () {
        if (isFiltered()) {
            return title.numIssues (this);
        } else {
            return title.numIssues (null);
        }
    }

    public Issue getIssueByArrayPos (int issueOffset) {
        if (isFiltered()) {
            return title.getIssueByArrayPos (issueOffset, this);
        } else {
            return title.getIssueByArrayPos (issueOffset, null);
        }
    }

    public double getGradeValue () {
        if (isFiltered()) {
            return title.getGradeValue (this);
        } else {
            return title.getGradeValue (null);
        }
    }

    /*
     * Some utility routines.
     */
    public Issue getLastIssue () {
        if (isFiltered()) {
            int numIssues = title.numIssues (this);
            return title.getIssueByArrayPos (numIssues-1, this);

        } else {
            int numIssues = title.numIssues (null);
            return title.getIssueByArrayPos (numIssues-1, null);
        }
    }
}