/******************************************************************************
 * Module Name : $RCSfile: AbstractReport.java,v $
 *
 * Description : Abstract report generator.
 *
 * Updates:
 *   $Log: AbstractReport.java,v $
 *   Revision 1.13  2002/07/13 01:06:20  jcouper
 *   job:157
 *   Add +/- for grades.  (Also fixed 147 at the same time).
 *
 *   Revision 1.12  2002/07/10 07:32:34  jcouper
 *   job:158
 *   Option to only report on held titles.
 *
 *   Revision 1.11  2001/10/11 20:32:02  jcouper
 *   146: Added WANT_COMPLETE
 *
 *   Revision 1.10  2001/10/08 22:10:52  jcouper
 *   145: Making issue codes smaller
 *
 *   Revision 1.9  2000/08/23 19:40:55  jcouper
 *   New doHeldOnly method for use, and a bit of tidying up.
 *
 *   Revision 1.8  2000/03/18 00:46:46  Jonathan
 *   Report enhancements to reduce size of long runs with the same grade.
 *
 *   Revision 1.7  2000/02/18 22:35:47  Jonathan
 *   Display reports in HTML.
 *
 *   Revision 1.6  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.5  1999/11/28 07:13:41  Jonathan
 *   Added Flags to Titles, and Polished AllTitlesMatrix report
 *
 *   Revision 1.4  1999/11/27 10:31:33  Jonathan
 *   Interface tweaks.  Reporting on the menu.  Fooling around with HTML
 *   about box.
 *
 *   Revision 1.3  1999/11/27 09:55:02  Jonathan
 *   Added a new report function.  Also some structural changes, create report
 *   objects rather than cheating with static methods.  Use Reflection to generate
 *   reports.
 *
 *   Revision 1.2  1999/11/15 23:22:15  Jonathan
 *   Rename for clarity
 *
 *   Revision 1.1  1999/11/14 04:32:24  Jonathan
 *   Basic reports
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.report;

import standup.parse.*;

public abstract class AbstractReport extends Object
{
    /*
     **************************************************************************
     * Global report parameters.
     **************************************************************************
     * These could be get/set, but we want to use them a lot, and they are
     * going to be modifiable by subclasses, so let's just take a gamble and
     * make the attributes.
     */
    protected static int tableWidthPixels = 640;
    protected static int rowHeightPixels = 24;
    protected static int numRowsPerPage = 40;
    protected static int numColsRow = 10;
    protected static int colWidthPercent = 100 / numColsRow;
    protected static int numColsTitle = 5;
    protected static int colTitleWidthPercent = 100 - ((numColsRow - numColsTitle) * colWidthPercent);
    protected static String unusedColor = "#F7F7F7";
    protected static String titleColor = "#DDDDDD";

    /*
     * We may trim to 'held only' reporting titles which have at least 21
     * issues, but which we hold less than 15% of the issues.  This gets
     * rid of things like Fantastic Four, or Avengers, which I don't collect,
     * but which I have a couple of scattered Spiderman Appearances.  I don't
     * want to see the whole lot in my printed report.
     */
    private static int heldOnlyMinCount = 11;
    private static int heldOnlyMaxPercent = 10; // Either...
    private static int heldOnlyMaxCount = 3;        // Or
    protected static boolean doHeldOnlyForTitle (int numTotal, int numHeld) {
        if (numTotal < heldOnlyMinCount) {
            return false;
        }
        if (numHeld <= heldOnlyMaxCount) {
            return true;
        }
        if ((numHeld * 100) <= (heldOnlyMaxPercent * numTotal)) {
            return true;
        }
        return false;
    }

    /*
     **************************************************************************
     * Our generate HTML function.  You should override this.
     **************************************************************************
     */
    public abstract StringBuffer generateHtml (Index index, Integer flavour);

    /*
     **************************************************************************
     * Default name for report class, please override.
     **************************************************************************
     */
    public String getDefaultName () {
        return "standup_report.html";
    }

    /*
     **************************************************************************
     * HTML fragment methods.
     **************************************************************************
     *
     * Methods returning full TD fragments.
     */
    protected String tagTrStart () {
        return "<TR HEIGHT=" + rowHeightPixels + ">\n";
    }
    protected String tagTrEnd () {
        return "</TR>\n";
    }
    protected String tagTdIssueHeld (Issue issue) {
        String code = issue.getCode();
        int underscore = code.indexOf ('_');

        // Form 100_ALT, or similar?
        if (Character.isDigit (code.charAt (0)) && (underscore != -1)) {
            code = "<FONT SIZE=-1>" + code.substring (0, underscore) + "</FONT><FONT SIZE=-3>" + code.substring (underscore + 1) + "</FONT>";
        } else {
            code = "<FONT SIZE=-1>" + code + "</FONT>";
        }

        if (issue.isHeld()) {
            return "<TD WIDTH=" + colWidthPercent + "%>" + code + " <FONT SIZE=-3>" + issue.getGradeAndModAbbreviation() + "</FONT></TD>";
        } else {
            return "<TD WIDTH=" + colWidthPercent + "%><I>" + code + "</I></TD>";
        }
    }
    protected String tagTdIssuesHeld (Issue issue1, Issue issue2, int numCols) {
        if (issue1.getCode().equals (issue2.getCode())) {
            if (issue1.isHeld()) {
                return "<TD COLSPAN=" + numCols + " WIDTH=" + colWidthPercent + "%>" + issue1.getCode() + " <FONT SIZE=-3>" + issue1.getGradeAndModAbbreviation() + "</FONT></TD>";
            } else {
                return "<TD COLSPAN=" + numCols + " WIDTH=" + colWidthPercent + "%><I>" + issue1.getCode() + "</I></TD>";
            }

        } else {
            if (issue1.isHeld()) {
                return "<TD COLSPAN=" + numCols + " WIDTH=" + colWidthPercent + "%>" + issue1.getCode() + " .. " + issue2.getCode() + " <FONT SIZE=-3>" + issue1.getGradeAndModAbbreviation() + "</FONT></TD>";
            } else {
                return "<TD COLSPAN=" + numCols + " WIDTH=" + colWidthPercent + "%><I>" + issue1.getCode() + " .. " + issue2.getCode() + "</I></TD>";
            }
        }
    }
    protected String tagTdTitle (String titleString) {
        return "<TH COLSPAN=" + numColsTitle + " WIDTH=" + colTitleWidthPercent + "% BGCOLOR=" + titleColor + ">" + titleString + "</TH>\n";
    }

    protected String tagTdBlank () {
        return "<TD BGCOLOR=" + unusedColor + " WIDTH=" + colWidthPercent + "%>&nbsp;</TD>";
    }

    /*
     * Methods returning full TR or TH fragments.
     */
    protected String tagTrTitle (String titleString) {
        return tagTrStart() + "<TH COLSPAN=" + numColsRow + " WIDTH=100% BGCOLOR=" + titleColor + ">" + titleString + "</TH>" + tagTrEnd();
    }
    protected String tagTrIssueInfo (Issue issue) {
        String description = issue.getInfo();
        if ((description == null) || description.equals("")) {
            description = issue.getCode();
        }
        if (issue.isHeld()) {
            return tagTrStart() + "<TD COLSPAN=" + numColsRow + " WIDTH=100%>" + description + " <FONT SIZE=-3>" + issue.getGradeAndModAbbreviation() + "</FONT></TD>" + tagTrEnd();
        } else {
            return tagTrStart() + "<TD COLSPAN=" + numColsRow + " WIDTH=100%><I>" + description + "</I></TD>" + tagTrEnd();
        }
    }

    /*
     * Other misc HTML fragments which we want to keep consistent.
     */
    protected String tagHead (String headString) {
        return "<HEAD><TITLE>" + headString + "</TITLE></HEAD>\n";
    }
    protected String tagH1 (String h1String) {
        return "<H1>" + h1String + "</H1>\n";
    }
    protected String tagHtmlStart () {
        return "<HTML>\n";
    }
    protected String tagHtmlEnd () {
        return "</HTML>\n";
    }
    protected String tagBodyStart () {
        return "<BODY BGCOLOR=#FFFFFF>\n";
    }
    protected String tagBodyEnd () {
        return "</BODY>\n";
    }
    protected String tagTableStart () {
        return "<CENTER><TABLE WIDTH=" + tableWidthPixels + " BORDER=1 CELLSPACING=0 CELLPADDING=2>\n";
    }
    protected String tagTableEnd () {
        return "</TABLE></CENTER>\n";
    }

    /*
     **************************************************************************
     * Other report utility methods.
     **************************************************************************
     */
    /*
     * Break table for new page if required.
     */
    protected boolean breakTableOnPage (int rowCount, int rowsNeeded, StringBuffer buf) {
        if ((rowCount + rowsNeeded) > numRowsPerPage) {
            buf.append (tagTableEnd());
            buf.append ("<br>");
            buf.append (tagTableStart());
            return true;
        }
        return false;
    }

    protected String replace (String string, String oldSub, String newSub) {
        String result = string;
        int pos = 0;
        while ((pos = result.indexOf (oldSub)) >= 0) {
            result = result.substring (0, pos) +
                     newSub +
                     result.substring (pos + oldSub.length());
        }
        return result;
    }

    protected String abbreviateTitle (String titleName) {
        String result = titleName;
        result = replace (result, "Limited Series", "L.S.");
        result = replace (result, "Annual", "Ann.");
        result = replace (result, "Volume", "V");
        result = replace (result, "Vol", "V");
        return result;
    }
}