/******************************************************************************
 * Module Name : $RCSfile: AllTitleMatrix.java,v $
 *
 * Description : Generates matrix report of all titles in an index.
 *
 * Updates:
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.report;

import java.util.*;
import standup.parse.*;

public class AllTitleMatrix extends AbstractReport
{
    // Keys for our report flavours
    public static final int TITLE_MATRIX = 0;

    public String getDefaultName () {
        return "title_matrix.html";
    }

    public StringBuffer generateHtml (Index index, Integer flavour) {
        StringBuffer buf = new StringBuffer ();
        Vector allTitles = index.getAllTitles();
        int totalIssues = 0;
        int totalIssuesHeld = 0;
        int rowCount = 0;

        index.sortAllTitlesByGetAnyName();

        buf.append (tagHtmlStart());
        buf.append (tagHead(index.toString()));
        buf.append (tagBodyStart());
        buf.append (tagTableStart());

       for (Enumeration e = allTitles.elements() ; e.hasMoreElements() ;) {
            Title title = (Title) e.nextElement();

            String titleName = title.getDefaultTitleName();

            /*
             * Get the number of issues, with no filtering, held and total.
             */
            TitleRef heldTitleRef = new TitleRef (index, title);
            heldTitleRef.setFilterHeldOnly(true);

            int numIssuesHeld = title.numIssues (heldTitleRef);
            int numIssuesTotal = title.numIssues ();
            int numIssues;
            boolean doHeldOnly;

            doHeldOnly = doHeldOnlyForTitle (numIssuesTotal, numIssuesHeld);
            numIssues = doHeldOnly ? numIssuesHeld : numIssuesTotal;

            /*
             * Case 0:
             * We have no issues?  e.g. we're doing held only, and have none.
             */
            if (numIssues == 0) {
                continue;

            /*
             * Case 1:
             * Is this title a funny one with a whole heap of non-numeric title
             * abbreviations, with the actual comic name held in the info field?
             */
            } else if (title.hasFlag (Title.flagInfoIsCode)) {
                // Need at least two rows left on this page.
                if (breakTableOnPage (rowCount, 2, buf)) {
                    rowCount = 0;
                }

                // Title gets a dedicated row.
                buf.append (tagTrTitle (titleName));
                rowCount++;

                // Then start running rows for the issues.
                for (int i=0; i<numIssuesTotal; i++) {
                    Issue issue = title.getIssueByArrayPos (i);
                    if (doHeldOnly && !issue.isHeld()) {
                        continue;
                    }

                    // Break table and continue on new page if required.
                    if (breakTableOnPage (rowCount, 1, buf)) {
                        buf.append (tagTrTitle ("... " + titleName + " (cntnd)"));
                        rowCount = 1;
                    }

                    // Now the actual tag.
                    buf.append (tagTrIssueInfo (issue) + "\n");
                    rowCount++;
                }

            /*
             * Case 2:
             * If there's few enough issue, we can pop it all on one line.
             */
            } else if (numIssues <= (numColsRow - numColsTitle)) {
                // Start new page if we're full.
                if (breakTableOnPage (rowCount, 1, buf)) {
                    rowCount = 0;
                }

                buf.append (tagTrStart ());
                buf.append (tagTdTitle (abbreviateTitle (titleName)));

                for (int i=0; i<numIssuesTotal; i++) {
                    Issue issue = title.getIssueByArrayPos (i);
                    if (doHeldOnly && !issue.isHeld()) {
                        continue;
                    }

                    buf.append (tagTdIssueHeld (issue) + "\n");
                }
                for (int i=0; i<(numColsRow - (numColsTitle + numIssues)); i++) {
                    buf.append (tagTdBlank() + "\n");
                }
                buf.append (tagTrEnd ());
                rowCount++;

            /*
             * Case 3:
             * Do we have a single range of multiple titles, all the same grade?
             * This is real simple.
             */
            } else if (title.consecutiveGradeFromPos (0) == numIssuesTotal) {
                // Start new page if we're full.
                if (breakTableOnPage (rowCount, 1, buf)) {
                    rowCount = 0;
                }

                buf.append (tagTrStart ());
                buf.append (tagTdTitle (titleName));

                Issue issue1 = title.getIssueByArrayPos (0);
                Issue issue2 = title.getIssueByArrayPos (numIssuesTotal-1);

                buf.append (tagTdIssuesHeld (issue1, issue2, numColsRow - numColsTitle));

                buf.append (tagTrEnd ());
                rowCount++;

            /*
             * Case 4:
             * If there are too many, put the title on its own line, and
             * then do the issues.  Most titles fall into this category.
             */
            } else {
                int columnCount = 0;

                // Need at least two rows left on this page.
                if (breakTableOnPage (rowCount, 2, buf)) {
                    rowCount = 0;
                }

                // Title gets a dedicated row.
                buf.append (tagTrTitle (titleName));
                rowCount++;

                // Then start running rows for the issues.
                buf.append (tagTrStart ());
                for (int i=0; i<numIssuesTotal; i++) {
                    Issue issue = title.getIssueByArrayPos (i);
                    if (doHeldOnly && !issue.isHeld()) {
                        continue;
                    }

                    // Start new column if we've finished this one.
                    if (columnCount >= numColsRow) {
                        buf.append (tagTrEnd ());
                        rowCount++;

                        // Break table and continue on new page if required.
                        if (breakTableOnPage (rowCount, 1, buf)) {
                            buf.append (tagTrTitle ("... " + titleName + " (cntnd)"));
                            rowCount = 1;
                        }

                        buf.append (tagTrStart ());
                        columnCount = 0;
                    }

                    // Have we got a clear run, and we're at the start of a row,
                    // in this case, we group whole rows into a single row.  We
                    // take blocks of numColsRow (usually 10) unless this is the
                    // end of the range, in which case we go to the end.
                    if ((columnCount == 0) && !doHeldOnly) {
                        int numConsecutive = title.consecutiveGradeFromPos (i);

                        int numConsecutiveRows = numConsecutive / numColsRow;
                        if (numConsecutiveRows > 0) {
                            int endPos = i + (numConsecutiveRows * numColsRow) - 1;
                            Issue endIssue = title.getIssueByArrayPos (endPos);

                            buf.append (tagTdIssuesHeld (issue, endIssue, numColsRow) + "\n");
                            i = endPos;
                            columnCount = numColsRow + 1;
                            continue;
                        }
                    }

                    // Now the actual tag.
                    buf.append (tagTdIssueHeld (issue) + "\n");
                    columnCount++;
                }

                // Fill any remaining columns, to be tidy.
                for (int i=0; i<(numColsRow - columnCount); i++) {
                    buf.append (tagTdBlank () + "\n");
                }
                buf.append (tagTrEnd ());
                rowCount++;
            }
       }
        buf.append (tagTableEnd());
        buf.append (tagBodyEnd());
        buf.append (tagHtmlEnd());

        return buf;
    }
}