/******************************************************************************
 * Module Name : $RCSfile: CoverDateComponent.java,v $
 *
 * Description : Editor for CoverDate elements.
 *
 * Updates:
 *   $Log: CoverDateComponent.java,v $
 *   Revision 1.3  2001/07/29 00:47:33  jcouper
 *   job:137
 *   Added support for credits, and re-vamped the EditDialog.
 *
 *   Revision 1.2  2000/02/18 22:35:53  Jonathan
 *   Display reports in HTML.
 *
 *   Revision 1.1  2000/01/10 02:24:55  Jonathan
 *   New utility components to simplify edit dialog.
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.util;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import standup.*;
import standup.parse.*;
import standup.util.*;

public class CoverDateComponent
    extends JComponent
    implements ItemListener, DocumentListener
{
    static final int STEP_SEQUENCE = CoverDate.STEP_MAX + 1;
    static final int STEP_UNKNOWN = CoverDate.STEP_MAX + 2;
    static final String stepSequenceString = "<Sequential>";
    static final String stepUnknownString = "<Not Dated>";

    CoverDate coverDate = null;
    boolean configuring = false;

    TopToBottomLayout layout = new TopToBottomLayout ();
    JNumberTextField yearField = new JNumberTextField (4);

    // Note that there are 13 month strings - month 13 is no-month (year only).
    JComboBox monthCombo = new JComboBox (CoverDate.monthCodes);
    JComboBox stepCombo = new JComboBox (CoverDate.stepStrings);
    JPanel yearMonthPanel = new JPanel ();

    public CoverDateComponent () {
        yearField.setBounds (CoverDate.YEAR_MIN, CoverDate.YEAR_MAX);

        /*
         * Add two more special options to the step box.
         */
        stepCombo.addItem (stepSequenceString);
        stepCombo.addItem (stepUnknownString);

        yearMonthPanel.add (monthCombo, BorderLayout.CENTER);
        yearMonthPanel.add (yearField, BorderLayout.EAST);

        layout.setColumns (1, 1);
        layout.setAlignment (TopToBottomLayout.ALIGN_LEFT);
        layout.setVAlignment (TopToBottomLayout.ALIGN_VTOP);
        layout.setVgap (5);
        setLayout (layout);
        add (new JLabel ("Date Sequence"));
        add (stepCombo);
        add (new JLabel ("Month/Year"));
        add (yearMonthPanel);

        setEnabled (false);

        yearField.getDocument().addDocumentListener (this);
        monthCombo.addItemListener (this);
        stepCombo.addItemListener (this);

        setSize (200, 40);
        configure ();
    }

    public CoverDateComponent (CoverDate inCoverDate) {
        this ();

        setCoverDate (inCoverDate);

        configure ();
    }

    public void setCoverDate (CoverDate inCoverDate) {
        coverDate = inCoverDate;

        configure ();
    }

    protected void configure () {
        configuring = true;

        if (coverDate == null) {
            setEnabled (false);

            yearField.setText ("1000");
            monthCombo.setSelectedIndex (0);
            stepCombo.setSelectedIndex (STEP_UNKNOWN);

        } else {
            setEnabled (true);
            stepCombo.setEnabled (true);

            yearField.setText (String.valueOf (coverDate.getYear()));
            monthCombo.setSelectedIndex (coverDate.getMonth() - 1);

            if (!coverDate.isDefined ()) {
                stepCombo.setSelectedIndex (STEP_UNKNOWN);
                yearField.setEnabled (false);
                monthCombo.setEnabled (false);

            } else if (!coverDate.isMark ()) {
                stepCombo.setSelectedIndex (STEP_SEQUENCE);
                yearField.setEnabled (false);
                monthCombo.setEnabled (false);

            } else {
                stepCombo.setSelectedIndex (coverDate.getStep());
                yearField.setEnabled (true);
                monthCombo.setEnabled (true);
            }
        }
        configuring = false;
    }

    /*
     **************************************************************************
     * LISTENER FUNCTIONS
     **************************************************************************
     */
    /**
     * adds an ActionListener to the button
     */
    public void addActionListener(ActionListener l) {
        listenerList.add(ActionListener.class, l);
    }

    /**
     * removes an ActionListener from the button
     */
    public void removeActionListener(ActionListener l) {
        listenerList.remove(ActionListener.class, l);
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireActionPerformed() {
        ActionEvent event = new ActionEvent (this, ActionEvent.ACTION_PERFORMED, "Change", 0);
        // Guaranteed to return a non-null array
        Object[] listeners = listenerList.getListenerList();
        // Process the listeners last to first, notifying
        // those that are interested in this event
        for (int i = listeners.length-2; i>=0; i-=2) {
            ((ActionListener)listeners[i+1]).actionPerformed(event);
        }
    }

    /*
     * Combo box listeners.
     */
    public void itemStateChanged(ItemEvent event) {
        if (configuring) {
            return;
        }

        Object source = event.getSource ();
        if (source == monthCombo) {
            coverDate.setMonth (monthCombo.getSelectedIndex () + 1);

            fireActionPerformed ();

        } else if (source == stepCombo) {
            int step = stepCombo.getSelectedIndex ();

            /*
             * Note that if we're messing around with cover dates,
             * the table should be being recomputed.
             */
            if (step == STEP_UNKNOWN) {
                coverDate.setMark (true);
                coverDate.setDefined (false);

            } else if (step == STEP_SEQUENCE) {
                coverDate.setMark (false);
                coverDate.setDefined (true);

            } else {
                coverDate.setMark (true);
                coverDate.setDefined (true);
                coverDate.setStep (step);

                if (monthCombo.getSelectedIndex() < 0) {
                    monthCombo.setSelectedIndex (0);
                }
                if (yearField.getInt() < CoverDate.YEAR_MIN) {
                    yearField.setText (String.valueOf (CoverDate.YEAR_MIN));
                }
            }

            configure ();
            fireActionPerformed ();
        }
    }

    /*
     * Act on document changes.
     */
    public void changedUpdate (DocumentEvent e) {
        documentUpdate (e);
    }
    public void insertUpdate (DocumentEvent e) {
        documentUpdate (e);
    }
    public void removeUpdate (DocumentEvent e) {
        documentUpdate (e);
    }
    public void documentUpdate (DocumentEvent e) {
        if (configuring) {
            return;
        }
        coverDate.setYear (yearField.getInt ());
        fireActionPerformed ();
    }
}