/******************************************************************************
 * Module Name : $RCSfile: LimitedLengthDocument.java,v $
 *
 * Description : Utility class.
 *
 * Updates:
 *   $Log: LimitedLengthDocument.java,v $
 *   Revision 1.2  1999/10/24 22:37:56  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends. 
 * 	      	  You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/ 
package standup.util;

import java.lang.Math;
import java.awt.Toolkit;
import javax.swing.text.PlainDocument;
import javax.swing.text.Document;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import javax.swing.text.JTextComponent;
	
/**
*	Document class which can be plugged into any JTextComponent in
*	order to give it the ability to length-limit its text.
*
*	@author	Peter Ashford
*
*/
public class LimitedLengthDocument extends PlainDocument{
	private int 			maxLength;	// the length text is limited to
	private int 			docLength;	// current length of text
	private JTextComponent	component;	// the component which is displaying the text
	
		
	/**
	*	Default constructor.  Creates a non length-limited document.  The 
	*	user needs to call the init() method in order to set the length
	*	limits for the document.
	*/
	public LimitedLengthDocument(){
		super();
		addDocumentListener(new limitedDocListener());
	
	}
	
	/**
	*	Sets the length limits for the document.
	*
	*	@param	textComponent	the handle of the JTextComponent to which this
	*							document is being added.
	*	@param	length			the length that the document is limited to.
	*/
	public void init (JTextComponent textComponent, int length){
		maxLength = length;
		component = textComponent;
	}
		
	//-------------------------------------------------------------
	public void insertString(int offs, String str, AttributeSet a) 
		throws BadLocationException {			
			
		if (str == null) {
			return;
		}  
		int slen = str.length();
		if(docLength + slen <= maxLength){
			super.insertString(offs, str, a);	
		}
		else{
			// extent is how many characters to copy into the string	
			int extent = Math.min(slen, maxLength-offs);

			// overflow is how many pre-existing characters would flow over 
			//		the bounds if we insert in the middle of existing text
			if(slen+docLength > maxLength){
				int overflow = 
					(Math.min(docLength, 
							  (offs+extent)-(maxLength-docLength))
					) - offs;
									
				// remove any text that will overflow	
				if(overflow>0){	
					int caret = component.getCaretPosition();
					component.select(docLength-overflow, docLength);
					component.replaceSelection(null);
					component.setCaretPosition(caret);
				}				
			}
			else{
				extent = slen;
			}

			Toolkit.getDefaultToolkit().beep();
			super.insertString(offs, str.substring(0, extent), a);
		}				
	}
	
	
	//=============================================================
	class limitedDocListener implements DocumentListener{
		public void insertUpdate(DocumentEvent e){
			docLength += e.getLength();
		}
		
		public void removeUpdate(DocumentEvent e){
			docLength -= e.getLength();
		}
		
		public void changedUpdate(DocumentEvent e){}	
	} // limitedDocListener
}