/******************************************************************************
 * Module Name : $RCSfile: NumberDocument.java,v $
 *
 * Description : Utility class.
 *
 * Updates:
 *   $Log: NumberDocument.java,v $
 *   Revision 1.5  2002/07/08 07:19:11  jcouper
 *   job:150
 *   Still fine tuning the entry of digits - making it more forgiving.
 *
 *   Revision 1.4  2002/07/07 02:45:50  jcouper
 *   job:156
 *   Cut & Paste for comic value field.
 *
 *   Revision 1.3  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.2  1999/10/24 22:37:56  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends.
 *            You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/
package standup.util;

import java.lang.Math;
import java.awt.Toolkit;
import javax.swing.text.JTextComponent;
import javax.swing.text.PlainDocument;
import javax.swing.text.Document;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;

/**
*   Document class which can be plugged into any JTextComponent in
*   order to give it the ability to limit its text to accept digits only.
*   May also be optionally length limited (number of characters) and/or bounds
*   limited (numerical range).
*
*   @author Peter Ashford
*/
public class NumberDocument extends PlainDocument {
    long min, max;
    int maxLength;
    boolean bounds_checking  = false;
    boolean length_checking  = false;
    boolean cents = false;

    /**
    *   Sets the minimum and maximum allowed decimal value for this field,
    *   and turns bounds checking on.
    *
    *   @param  minVal  the minimum value to accept
    *   @param  maxVal  the maximum value to accept (>minVal)
    *   @param  text    the text component that this document belongs to - we
    *                   need this to check when focus is lost so that minimum
    *                   values can be set if input is less than minimum.
    */
    public void setMinMax(long minVal, long maxVal, JTextComponent text){
        min = minVal;
        max = maxVal;
        text.addFocusListener(new textFocusL());
        bounds_checking = true;

        // make sure initial/existing value is ok
        try{
            String curText = getText(0, getLength());

            if(curText!=null && curText.length()>0){
                long val = Long.parseLong(curText);
                if(val<min){
                    remove(0,getLength());
                    super.insertString(0, String.valueOf(min), null);
                } else if(val>max){
                    remove(0,getLength());
                    super.insertString(0, String.valueOf(max), null);
                }
            } else {
                remove(0,getLength());
                super.insertString(0, String.valueOf(min), null);
            }
        }
        catch(Exception e){
        }
    }


    /**
    *   Sets the "00 vs "0" flag.
    *
    *   @param  value   true or false.
    */
    public void setCents(boolean value){
        cents = value;
    }


    /**
    *   Sets the maximum number of characters to accept in this field and.
    *   turns length checking on.
    *
    *   @param  maxLen  maximum number of characters to limit the field to.
    */
    public void setMaxLength(int maxLen){
        maxLength       = maxLen;
        length_checking = true;
    }


    /**
    *   Here to allow textFocusL to access the super class's InsertString
    *   method.
    */
    protected void superInsertString(String str){
        try{
            super.insertString(0, str, null);
        }
        catch(Exception e){
        }
    }


    /**
    *    Listen to the this text document attached to the component.
    *    If the component loses focus and the text field is null or empty
    *    or a lone minus sign, then set the text field to "0" or "00".
    */
    class textFocusL extends FocusAdapter {
        public void focusLost (FocusEvent fe) {
            try {
                String curText = getText (0, getLength ());
                if (curText == null || curText.length () == 0 ||
                    (curText.length () == 1 && curText.charAt (0) == '-') ||
                    (curText.length () == 1 && curText.charAt (0) == '0')) {

                    remove (0, getLength ());
                    if (cents) {
                        superInsertString ("00");
                    } else {
                        superInsertString ("0");
                    }
                }
            }
            catch (Exception e) {
            }
        }
    }



    /**
    *   We simply take input strings and only insert the numeric content
    *   into the document
    */
    public void insertString(int offs, String str, AttributeSet a)
        throws BadLocationException {

        if (str == null || str.length()==0) {
            return;
        }

        char[] digits = str.toCharArray();
        int j=0;

        // strip non-digit characters from digits[] array
        for (int i=0; i < digits.length; i++) {
            if(Character.isDigit(digits[i]) || (digits[i] == '-' && i == 0)){
                digits[j++] = digits[i];
            }
            else{
                Toolkit.getDefaultToolkit().beep();
            }
        }


        if(j==0){
            return;
        }

        // perform length checking.
        if(length_checking){
            int l = str.length() + getLength();
            int diff = l-maxLength;

            if(diff>0){
                Toolkit.getDefaultToolkit().beep();
                return;
            }
        }

        // perform bounds-checking.  We  calculate what value the insertion
        // will make our current text, and either insert it if it's ok, or
        // set it to max if it overflows bounds.
        if(bounds_checking){
            int negCount = 0;
            str = new String(digits,0,j);
            String newText = getText(0,offs);
            newText += str;
            newText += getText(offs, getLength()-offs);

            // error filtering code for negative numbers...
            for (int i = 0; i < newText.length (); i++) {
                if (newText.charAt (i) == '-') {
                    negCount = negCount + 1;
                }
            }

            switch (negCount) {
                case 0: // choke if no minus char but can only be negative
                    if (min < 0 && max < 0) {
                        Toolkit.getDefaultToolkit ().beep ();
                        return;
                    }
                    break;
                case 1: // must be positive or have a minus sign only at the head
                    if ((newText.charAt (0) != '-') || (min >= 0 && max >= 0)) {
                        Toolkit.getDefaultToolkit ().beep ();
                        return;
                    }
                    break;
                default: // choke if more than 1 minus char
                    Toolkit.getDefaultToolkit ().beep ();
                    return;
            }

            // Handle negative numbers...
            if (newText.charAt (0) == '-') {
                switch (newText.length ()) {
                case 1:    // insert the leading minus sign
                    super.insertString(offs, new String(digits,0,digits.length), a);
                    break;
                case 2:    // allow insert just after the minus only if not zero
                    if (newText.charAt (1) != '0') {
                        super.insertString(offs, new String(digits,0,digits.length), a);
                    }
                    break;
                default:   // the usual stuff, but for negative numbers...
                    long val = Long.parseLong(newText);

                    if (val < min) {
                        // value entered will underflow minimum - set value
                        // to min allowed instead
                        remove (0, getLength ());
                        super.insertString(0, String.valueOf(min), a);
                    }
                    else{
                        super.insertString(offs, new String(digits,0,digits.length), a);
                    }
                } // switch (newText.length ())
            } // if (newText.charAt (0) == '-')

            // Handle positive numbers...
            else {

                /*
                 * If value entered will overflow maximum - set value
                 * to max allowed instead.
                 */
                long val = Long.parseLong (newText);
                if (val > max){
                    remove (0, getLength());
                    super.insertString (0, String.valueOf (max), a);
                    return;
                }

                /*
                 * no leading zeroes allowed on multidigit numbers!
                 * (except for cents case, which allows one)
                 */
                int maxLeadZeroLen = (cents ? 2 : 1);
                if ((newText.length() > maxLeadZeroLen) && (newText.charAt (0) == '0')) {
                    while ((newText.length() > maxLeadZeroLen) && (newText.charAt (0) == '0')) {
                        newText = newText.substring (1);
                    }
                    remove (0, getLength());
                    super.insertString (0, newText, a);
                    return;
                }

                /* No problems... insert the digits! */
                super.insertString (offs, new String (digits, 0, digits.length), a);
            }
        } // if(bounds_checking)
        else{
            super.insertString(offs, new String(digits,0,digits.length), a);
        }
    } // public void insertString(

} // class NumberDocument