/******************************************************************************
 * Module Name : $RCSfile: TopToBottomLayout.java,v $
 *
 * Description : Utility class.
 *
 * Updates:
 *   $Log: TopToBottomLayout.java,v $
 *   Revision 1.3  2000/01/09 00:11:55  Jonathan
 *   Edit Dialog, Preferences
 *
 *   Revision 1.2  1999/10/24 22:37:56  Jonathan
 *   Proper CVS headers.  Fix A-018/19/20
 *
 *
 * Copyright Notice:
 * (c) 1999. This source code was developed by Jonathan Couper and friends. 
 * 	      	  You may use it for any purpose.  No guarantees given or implied.
 *****************************************************************************/ 
package standup.util;

import java.awt.*;
import java.awt.event.*;


/**
*   Layout a resizable, speadsheet-like component in left-to-right, top-to-bottom 
*   fashion. This manager attempts to set the columns and rows according to the
*   largest preferred size in each column and row. If the layout is re-sized 
*   smaller than the preferred minima of the columns and rows, a truncated visual
*   display shall result, dictated by the edges of the parent component. If the 
*   layout is re-sized larger than the preferred minima of the sub-components, a 
*   designated column (and possibly a row) are dynamically stretched to fill out
*   to the parent's dimensions.
*/
public class TopToBottomLayout
implements LayoutManager2 {

    /**
    *   Align each subcomponent to the left side of their display cell.
    */
    public static final int ALIGN_LEFT = 0;

    /**
    *   Align each subcomponent to the right side of their display cell.
    */
    public static final int ALIGN_RIGHT = 1;

    /**
    *   Align each subcomponent in the centre of their display cell.
    */
    public static final int ALIGN_CENTER = 2;

    /**
    *   Same as ALIGN_LEFT.
    */
    public static final int ALIGN_STRETCH = 3;


    /**
    *   Disallow the slackRow to be re-sized.
    */
    public static final int ALIGN_VNONE = 4;

    /**
    *   Allow the slackRow to be top-aligned.
    */
    public static final int ALIGN_VSTRETCH = 5;

    /**
    *   Allow the slackRow to be top-aligned.
    */
    public static final int ALIGN_VTOP = 6;

    /**
    *   Allow the slackRow to be re-sized.
    */
    public static final int ALIGN_VCENTER = 7;

    /**
    *   Allow the slackRow to be bottom-aligned.
    */
    public static final int ALIGN_VBOTTOM = 8;


    private int vGap        = 0;
    private int hGap        = 0;
    private int alignment   = ALIGN_STRETCH;
    private int v_alignment = ALIGN_VNONE;
    private int numColumns  = 1;
    private int numRows     = 0;
    private int slackColumn = 1;
    private int slackRow    = 0;

    private int totalHeight;
    private int totalWidth;
    private Insets margins;
    private int columnWidth [];
    private int rowHeight [];


    /**
    *   Construct a TopToBottom layout with default alignment of ALIGN_STRETCH
    */
    public TopToBottomLayout () {
        setSlackRow (-1);
    }

    /**
    *   Construct a TopToBottom layout with the specified alignment.
    *
    *   @params alignment - an ALIGN_ constant belonging to this class.
    */
    public TopToBottomLayout (int alignmentIn) {
        super ();
        setAlignment (alignmentIn);
    }

    /**
    *   Change the layout's column alignment.
    *
    *   @params alignment - an ALIGN_ constant belonging to this class.
    */
    public void setAlignment (int alignmentIn) {
        alignment = alignmentIn;
    }

    /**
    *   Change the layout's column and row alignments.
    *
    *   @params alignment - an ALIGN_ constant belonging to this class.
    */
    public void setAlignment (int hAlignmentIn, int vAlignmentIn) {
        alignment = hAlignmentIn;
        setVAlignment (vAlignmentIn);
    }

    /**
    *   Change the layout's row alignment.
    *
    *   @params alignment - an ALIGN_ constant belonging to this class.
    */
    public void setVAlignment (int alignmentIn) {
        v_alignment = alignmentIn;
    }

    /**
    *   Set gap between rows.
    *
    *   @params number of pixels.
    */
    public void setVgap (int value) {
        vGap = value;
    }

    /**
    *   Set gap between columns.
    *
    *   @params number of pixels.
    */
    public void setHgap (int value) {
        hGap = value;
    }

    /**
    *   Set the row (origin 0) that shall re-size to fill expanded parent height.
    *
    *   @params rowNumberIn
    */
    public void setSlackRow (int rowNumberIn) {
        if (rowNumberIn >= 0) {
            v_alignment = ALIGN_VSTRETCH;
            numRows = rowNumberIn;

            if ((rowNumberIn >= 0) && (rowNumberIn <= numRows)) {
                slackRow = rowNumberIn;
            }
            else {
                slackRow = numRows;
            }
        }
        else {
            v_alignment = ALIGN_VNONE;
        }
    }

    /**
    *   Set the expected number of columns and the resizable slack column.
    *
    *   @params number of columns per row.
    *   @params column number which takes the slack for re-sizing.
    */
    public void setColumns (int numColumnsIn, int slackColumnIn) {
        if (numColumnsIn > 0) {
            numColumns = numColumnsIn;

            if ((slackColumnIn >= 1) && (slackColumnIn <= numColumns)) {
                slackColumn = slackColumnIn;
            }
            else {
                slackColumn = numColumns;
            }
        }
    }

    public void addLayoutComponent (String name, Component comp) {
    }

    public void removeLayoutComponent (Component comp) {
    }

    public Dimension preferredLayoutSize (Container parent) {

        int numComps    = parent.getComponentCount ();
        int column      = 0;
        int row         = 0;
        int rHeight     = 0;

        numRows         = numComps / numColumns;
        columnWidth     = new int [numColumns];
        rowHeight       = new int [numRows];
        margins         = parent.getInsets();
        totalHeight     = 0;

        //  Compute the preferred total column widths and total row height.

        for (int a = 0; a < numComps; a++) {
            Component c;
            Dimension d;

            c = parent.getComponent (a);
            if (c.isVisible ()) {
                d = c.getPreferredSize ();

                if (d.width > columnWidth [column]) {
                    columnWidth [column] = d.width;
                }

                if (rowHeight [row] < d.height) {
                    rowHeight [row] = d.height;
                }

                column = column + 1;
                if ((column == numColumns) || (a == (numComps - 1))) {
                    totalHeight = totalHeight + rowHeight [row];
                    row = row + 1;
                    column = 0;
                }
            }
        }

        //  Compute the total height.
        totalHeight = totalDimension (rowHeight, vGap) + margins.top + margins.bottom;

        //  Compute the total width.
        totalWidth = totalDimension (columnWidth, hGap) + margins.left + margins.right;

        return new Dimension (totalWidth, totalHeight);
    }


    /**
    *   Compute the total size (in pixels) of the given corrdinate array.
    *
    *   @params coord array reference.
    *   @params gap to insert between consecutive coords.
    */
    private int totalDimension (int [] coord, int gap) {

        int    total = 0;
        int    cardinality = coord.length;

        if (cardinality > 1) {
            total = (cardinality - 1) * gap;
        }

        for (int a = 0; a < cardinality; a++) {
            total = total + coord [a];
        }

        return total;
    }

    public Dimension minimumLayoutSize(Container parent) {
        return preferredLayoutSize (parent);
    }


    public void layoutContainer (Container parent) {

        int numComps    = parent.getComponentCount ();
        int column      = 0;
        int row         = 0;
        int rHeight     = 0;

        // Compute visible column widths.
        preferredLayoutSize (parent);
        int y           = margins.top;

        // Resize the designated slack column if the parent's width is not completely filled
        int parentWidth = parent.getSize ().width;
        if (totalWidth < parentWidth) {
            columnWidth [slackColumn - 1] =
                columnWidth [slackColumn - 1] + (parentWidth - totalWidth);
        }

        // Resize the designated slack row if the parent's height is not completely filled
        int parentHeight = parent.getSize ().height;
        if (totalHeight < parentHeight) {
            rowHeight [slackRow] =
                rowHeight [slackRow] + (parentHeight - totalHeight);
        }

        // Set the location, width and height of each component.
        for (int a = 0; a < numComps; a++) {
            Component c;
            Dimension d;

            c = parent.getComponent (a);
            if (c.isVisible ()) {
                d = c.getPreferredSize ();

                if (alignment == ALIGN_STRETCH) {
                    d.width = columnWidth [column];
                }

                if (v_alignment == ALIGN_VSTRETCH) {
                    d.height = rowHeight [row];
                }
                c.setSize (d);

                int x = margins.left;
                for (int i = 0; i < column; i++) {
                    x = x + columnWidth [i] + hGap;
                }

                if ((column == 0) && (a > 0)) {
                    y = y + vGap;
                }

                switch (alignment) {
                case ALIGN_RIGHT:
                    x = x + (columnWidth [column] - d.width);
                    break;
                case ALIGN_CENTER:
                    x = x + (columnWidth [column] - d.width) / 2;
                    break;
                default: // ALIGN_STRETCH & ALIGN_LEFT
                }

//                if (newBehaviour == true) {
//                    switch (v_alignment) {
//                    case ALIGN_VBOTTOM:
//                        y = y + (rowHeight [row] - d.height);
//                        break;
//                    case ALIGN_VCENTER:
//                        y = y + (rowHeight [row] - d.height) / 2;
//                        break;
//                    default: // ALIGN_VSTRETCH & ALIGN_VTOP
//                    }
//                }

                c.setLocation (x, y);

                column = column + 1;
                if ((column == numColumns) || (a == (numComps - 1))) {
                    y = y + d.height;
                    row = row + 1;
                    column = 0;
                }
            }
        }
    }


    public void addLayoutComponent(Component c, Object constraints) {
    }

    public Dimension maximumLayoutSize(Container target) {
        return null;
    }

    public float getLayoutAlignmentX(Container target) {
        return 0;
    }

    public float getLayoutAlignmentY(Container target) {
        return 0;
    }

    public void invalidateLayout(Container target) {
    }
}